#include <gtk/gtk.h>
#include <gtk/gtkenums.h>
#include <gtk/gtkrange.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <limits.h>
#include "playif.h"
#include "editif.h"
#include "editor.h"
#include "volume.h"
#include "../version.h"

#if 0
#define _DEBUG 1
#endif

#include "../portability.h"

#define IMAGE_PAUSE   "pause.xpm"
#define IMAGE_PLAY    "play.xpm"
#define IMAGE_PLAYTO  "next_t.xpm"
#define IMAGE_RECORD  "record.xpm"
#define IMAGE_STOP    "stop.xpm"
#define IMAGE_EJECT   "eject.xpm"

#define TIP_PLAYTO    "Play previous 5 seconds"
#define TIP_PAUSE     "Pause playback"
#define TIP_STOP      "Stop playback"
#define TIP_PLAY      "Play from current time"
#define TIP_RECORD    "Record current time"

#define EDIT_ACTION_COPY   1
#define EDIT_ACTION_DELETE 2
#define EDIT_ACTION_PASTE  3
#define EDIT_ACTION_INSERT 4
#define EDIT_ACTION_TIME   5
#define EDIT_ACTION_DUP    6

#define MAINMENU_ACTION_ABOUT          1
#define MAINMENU_ACTION_OPTIONS        2
#define MAINMENU_ACTION_LOAD_ABANDONED 3
#define MAINMENU_ACTION_OUTFILE        4
#define MAINMENU_ACTION_SAVE_ABANDON   5
#define MAINMENU_ACTION_EDIT           6
#define MAINMENU_ACTION_CLEAR          7
#define MAINMENU_ACTION_OPEN           8

#define OPTIONMENU_ACTION_STOP  1
#define OPTIONMENU_ACTION_PAUSE 2
#define OPTIONMENU_ACTION_VOLUME 3


#define OUTPUT_FILE_APPEND_DEFAULT 0
#define OUTPUT_FILE_APPEND_JOIN    1
#define OUTPUT_FILE_APPEND_SPLIT   2

#define EDIT_IDLE_STATE_INIT       0
#define EDIT_IDLE_STATE_LOOP       1
#define EDIT_IDLE_STATE_EDIT       2
#define EDIT_IDLE_STATE_DONE       3
#define EDIT_IDLE_STATE_ERROR      4

/* adam/TBD: only one index idleproc for now */
#define MAX_IDLE_INDEX_PROC 1
#define CONFIG_FILE_NAME ".xmpgeditrc"
#define GTK_RC_FILE      "/.xmpgedit-gtkrc"

#define XMPGEDIT_CREDITS_COPYRIGHT \
  "                    Credits\n" \
  "\nWritten by: Adam Bernstein <number6@mpgedit.org> under\n" \
  "the GPL.  See \"COPYING\" for full details.\n" \
  "\n" \
  "The mpgedit audio playback plugins uses the MPGLIB decoding engine,\n" \
  "from the MPG123 package, written by: Michael Hipp (www.mpg123.de)\n" \
  "MPGLIB is released under the GPL.\n" \
  "\n" \
  "The version of MPGLIB used in this release comes " \
  "from LAME 3.90 alpha 7,\n" \
  "released under the GNU LESSER PUBLIC LICENSE.\n" \
  "\n" \
  "Copyrights (c) 1999,2000,2001 by Mark Taylor: LAME\n" \
  "Copyrights (c) 1998 by Michael Cheng: LAME\n" \
  "Copyrights (c) 1995,1996,1997 by Michael Hipp: mpglib\n" \
  "\n" \
  "As well as additional copyrights as documented in the source code.\n"

typedef struct _time_spinbutton_ctx {
    GtkWidget *minute;
    GtkWidget *second;
    GtkWidget *millisecond;

    gdouble   minute_prev;
    gdouble   second_prev;
    gdouble   millisecond_prev;
} time_spinbutton_ctx;


/*
 * Function to call after ___row_modified_idleproc() has completed.
 */
typedef struct _row_modified_dialog_ctx {
    void      (*proc)(void *data);
    GtkWidget *widget;
    GdkEvent  *event;
    void      *data;
} row_modified_dialog_ctx;


typedef struct _outfile_dialog_ctx {
    struct _global_vars_ctx *gvctx;
    GtkWidget       *entry;
    GtkWidget       *dialog;
    GtkWidget       *radio_split;
    int             radio_selection;
    int             action;
    char            **file;
} outfile_dialog_ctx;


typedef struct _outfile_radio_ctx {
    int button;
    int *radio_selection;
} outfile_radio_ctx;


/*
 * Global variable context.  This context references both playback
 * controls, and the file selection widget.  Use this context
 * when access is needed to both playback controls, or one playback control 
 * needs access to the other.
 */
typedef struct _global_vars_ctx {
    GtkWidget *main_window;
    GtkWidget *playto_toggle;
    GtkWidget *pause_toggle;
    GtkWidget *play_toggle;
    GtkWidget *stop_toggle;
    GtkWidget *edit_menu;
    GtkWidget *main_menu;
    GtkWidget *about_dialog;
    GtkWidget *entry;
    GtkWidget  *open_file_toggle;

    /*
     * Widgets accessed after creation
     */
    GtkWidget *menu_edit_pbelow;
    GtkWidget *menu_save_session;
    GtkWidget *menu_edit_cut;
    GtkWidget *menu_edit_copy;
    GtkWidget *menu_edit_duplicate;
    GtkWidget *menu_edit_time;
    GtkWidget *menu_edit;

    GtkTooltips *tooltips;
    GtkTooltips *tooltips_entry;
    edit_ctx  editctx;
    volume_control_ctx *volumectx;
    outfile_dialog_ctx *ofdialogctx;
    outfile_dialog_ctx *abandon_file_ctx;
    outfile_dialog_ctx *query_abandon_ctx;

    /* File selection */
    GtkWidget *fs;

    /* Toolkit properties, needed for pixmap lookup */
    GtkSettings *settings;

    row_modified_dialog_ctx row_modified_ctx;

    /* 
     * Back reference to both playback controls.
     */
    struct _idle_ctx  *idlectx;
    struct _idle_ctx  *eidlectx;

    /*
     * File indexing parameters
     */
    int                     idle_indexfile_count;
    struct _index_file_list *ifl_head;
    struct _index_file_list *ifl_tail;

    int edit_action_abandon_save;
    /*
     * Configurable values
     */
    int pause_visable;
    int stop_visable;
    int volume_visable;
    char *output_file;
    char *abandon_file;
} global_vars_ctx;


typedef struct _edit_menu_ctx {
    global_vars_ctx *gvctx;
    int      action;
    char     *actionstr;
} edit_menu_ctx;


typedef struct _idle_ctx {
    GtkWidget           *label_elapsed;
    GtkWidget           *label_start;
    GtkWidget           *label_trt;
    GtkWidget           *scrollbar;
    GtkWidget           *play_toggle;
    GtkWidget           *playto_toggle;
    GtkWidget           *pause_toggle;
    GtkWidget           *stop_toggle;
    GtkWidget           *record;
    GtkObject           *adj;
    int                 adj_elapsed_label_noset;
    gint                idle_tag;
    void                *playctx;
    char                *play_file;
    int                 play_toggle_ignore;
    int                 ignore_spin_button_set;
    int                 ignore_adjustment_set;
    int                 ignore_spin_button_callbacks;
    
    int                 stime_sec;
    int                 stime_usec;

    int                 elapsed_sec;
    int                 elapsed_usec;

    int                 pause_sec;
    int                 pause_usec;

    int                 saved_sec;

    int                 playto_sec;
    int                 playto_msec;
    int                 dirty;
    int                 dirty_saved;
    int                 dirty_sec;
    int                 dirty_usec;
    edit_row_ctx        *dirty_row;
    time_spinbutton_ctx sbctx;
    global_vars_ctx     *gvctx;
} idle_ctx;


typedef struct _idle_indexfile_ctx {
    GtkWidget        *widget;
    GdkEvent         *event;
    gpointer         *data;
    GtkFileSelection *fs;
    void             *playctx;
    edit_row_ctx     *edit_row;
    long             tlast;
    long             fsize;
    char             *file_name;
    void             (*callback)(GtkWidget *, GdkEvent *, gpointer);
} idle_indexfile_ctx;

typedef struct _index_file_list {
    idle_indexfile_ctx      *ifctx;
    struct _index_file_list *next;
} index_file_list;



typedef struct _indexfile_idleproc_ctx {
    global_vars_ctx    *gvctx;
    idle_indexfile_ctx *ifctx;

    /* Progress bar controls */
    GtkWidget        *progress;
    GtkWidget        *dialog;
    GtkWidget        *label;
    int              quit;
} indexfile_idleproc_ctx;


/* 
 * Flow of control: cb_edit() -> cb_edit_idle_editfile()
 */
typedef struct _idle_editfile_ctx {
    global_vars_ctx *gvctx;
    void            *edfiles_ctx;
    int             edfiles_ctx_editing;
    GtkWidget       *dialog;
    GtkWidget       *progress;
    int             total_sec;
    int             current_sec;
    int             quit;
    int             edit_flags;
    int             state;
    int             edit_indx;
    int             out_fileindx;
    char            *out_filename;
    editspec_t      *edarray;
    int             edarray_len;
    editspec_t      *built_edarray;
} idle_editfile_ctx;


typedef struct _cb_edit_ctx
{
    int        index;
    int        len;
    editspec_t *edarray;
} cb_edit_ctx;


void ___file_sel_ok(GtkWidget *widget, gpointer *data,
                    long init_sec, long init_msec);

void  cb_row_clicked(GtkWidget *widget,
                     int row, int column,
                     GdkEventButton *event,
                     gpointer data);
void cb_edit_3(GtkWidget *widget, GdkEvent *dummy, gpointer data);
void cb_stop_button(GtkWidget *widget, gpointer data);
void cb_record_button(GtkWidget *widget, gpointer *data);
void ___set_editor_defaults(global_vars_ctx *gvctx);
void cb_open_file(GtkWidget *widget, gpointer data);
void cb_open_file_from_menu(gpointer data);
gint cb_quit(GtkWidget *widget, gpointer data);
void mainmenu_item_response(edit_menu_ctx *emctx);

void cb_edit_1(GtkWidget *widget, gpointer data);
void ___load_editor_by_files(global_vars_ctx *gvctx, char **files, int len);
void label_set_text(GtkWidget *widget, char *msg, ...);

GtkWidget *___make_indexfile_progress_dialog(indexfile_idleproc_ctx *ctx,
                                             GtkWidget *parent,
                                             char *file_name);
void cb_row_clicked_1(GtkWidget *widget, GdkEvent *event, gpointer data);
void ___spin_scrollbar_set_value(idle_ctx *idlectx);

#ifdef _DEBUG
void find_fd_leaks(char *label)
{
    char dummy_sb[1024];
    int i;

    for (i=0; i<2048; i++) {
        if (fstat(i, (void *) dummy_sb) == -1) {
            printf("find_fd_leaks: max_fd = %d %s\n", i, label);
            return;
        }
    }
}
#endif


int get_config_value(char *str, char *name, int *val)
{
    int len;
    char *cp = str;

    len = strlen(name);
    if (strncmp(cp, name, len)) {
        return 0;
    }

    cp += len;
    while (isspace(*cp)) cp++;
    if (*cp == ':') cp++;
    while (isspace(*cp)) cp++;
    
    if (!strcmp(cp, "TRUE")) {
        *val = 1;
    }
    else if (!strcmp(cp, "FALSE")) {
        *val = 0;
    }
    else {
        *val = atoi(cp);
    }
    return 1;
            
}


void write_config_file(char *file, global_vars_ctx *gvctx)
{
    FILE *fp    = NULL;

    fp = fopen(file, "w+");
    if (!fp) {
        goto clean_exit;
    }

    fprintf(fp, "pause_button: %s\n", gvctx->pause_visable ? "TRUE" : "FALSE");
    fprintf(fp, "stop_button:  %s\n", gvctx->stop_visable ?  "TRUE" : "FALSE");
    fprintf(fp, "volume_control: %s\n", 
            gvctx->volume_visable ?  "TRUE" : "FALSE");

clean_exit:
    if (fp) {
        fclose(fp);
    }
}


void read_config_file(char *file, global_vars_ctx *gvctx)
{
    FILE *fp    = NULL;
    char *buf   = NULL;
    int  buflen = 1024;
    char *cp;
    int  val;

    fp = fopen(file, "r");
    if (!fp) {
        goto clean_exit;
    }
    buf = malloc(buflen);
    if (!buf) {
        goto clean_exit;
    }

    do {
        cp = fgets(buf, buflen-1, fp);
        if (cp) {
            buf[strlen(buf)-1] = '\0';
            if (*cp == '#') {
                continue;
            }
            else if (get_config_value(cp, "pause_button:", &val)) {
                gvctx->pause_visable = val;
            }
            else if (get_config_value(cp, "stop_button:", &val)) {
                gvctx->stop_visable = val;
            }
            else if (get_config_value(cp, "volume_control:", &val)) {
                gvctx->volume_visable = val;
            }
        }
    } while (!feof(fp));

clean_exit:
    if (fp) {
        fclose(fp);
    }
    if (buf) {
        free(buf);
    }
    
}


void ___edit_row_set_focus(global_vars_ctx *gvctx)
{
    edit_row_ctx *er = gvctx->editctx.cursor;

#ifdef _DEBUG
    printf("Row %d selected: %s %d %d %d - %d %d %d\n", 
           er->row_num, er->file_name, 
           er->s_min, er->s_sec, er->s_msec,
           er->e_min, er->e_sec, er->e_msec);
#endif
    gvctx->idlectx->play_file  = strdup(er->file_name);
    gvctx->eidlectx->play_file = gvctx->idlectx->play_file;

    /*
     * Set entry file name
     */
    gtk_tooltips_enable(gvctx->tooltips_entry);
    gtk_tooltips_set_tip(gvctx->tooltips_entry, GTK_WIDGET(gvctx->entry),
                         er->file_name, NULL);
    gtk_entry_set_text(GTK_ENTRY(gvctx->entry), er->file_name);
    gtk_editable_set_position(
        GTK_EDITABLE(gvctx->entry), strlen(er->file_name));

    ___file_sel_ok(NULL, (gpointer) gvctx->idlectx,
                   er->s_min*60 + er->s_sec, er->s_msec);

    gvctx->eidlectx->adj_elapsed_label_noset = 1;
    ___file_sel_ok(NULL, (gpointer) gvctx->eidlectx,
                   er->e_min*60 + er->e_sec, er->e_msec);
    gvctx->eidlectx->adj_elapsed_label_noset = 0;

    /* 
     * This select_row call must be here.
     */
    gtk_clist_select_row(GTK_CLIST(gvctx->editctx.clist), er->row_num, 0); 
}



void ___gtk_box_add_index(GtkBox *box, GtkWidget *child, gint indx)
{
    gtk_box_pack_start(box, child, FALSE, FALSE, 0);
    gtk_box_reorder_child(box, child, indx);
    gtk_widget_show(child);
}


GtkWidget *___make_dialog_informational(GtkWidget *widget,
                                        gint cancel_button,
                                        char *file_name,
                                        char *msg)
{
    GtkWidget *dialog;
    GtkWidget *label;


    if (cancel_button) {
        dialog = gtk_dialog_new_with_buttons(
                     msg,
                     GTK_WINDOW(widget),
                     GTK_DIALOG_MODAL, 
                     GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
                     GTK_STOCK_OK,     GTK_RESPONSE_OK,
                     NULL);
    }
    else {
        dialog = gtk_dialog_new_with_buttons(
                     msg,
                     GTK_WINDOW(widget),
                     GTK_DIALOG_MODAL, 
                     GTK_STOCK_OK,     GTK_RESPONSE_OK,
                     NULL);
    }

    label = gtk_label_new(msg);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
                       label, FALSE, FALSE, 0);
    if (file_name) {
        label = gtk_label_new(file_name);
        gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
                           label, FALSE, FALSE, 0);
    }
    return dialog;
}


gint ___make_dialog_informational_run(global_vars_ctx *gvctx,
                                      gint cancel_button,
                                      char *msg)
{
    gint response;
    GtkWidget *dialog;

    dialog = ___make_dialog_informational(
                 gvctx->main_window,
                 cancel_button, NULL,
                 msg);
    gtk_widget_show_all(dialog);
    response = gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
    return response;
}


void ___make_cant_open_sound(global_vars_ctx *gvctx)
{
    static char *msg = 
      "                Couldn't Open Sound Device\n\n\n"
      "    Please verify that:\n"
      "      1. No other programs are using the soundcard      \n"
      "      2. Your soundcard is properly configured          \n";


    ___make_dialog_informational_run(gvctx, FALSE, msg);
}


void ___row_modified_dialog_1(global_vars_ctx *gvctx)
{
    gint response;
    char msg[128];
    char *waswere = "time was ";

    cb_stop_button(gvctx->idlectx->stop_toggle, gvctx->idlectx);
    cb_stop_button(gvctx->eidlectx->stop_toggle, gvctx->eidlectx);
    strcpy(msg, "        Current row ");
    if (gvctx->idlectx->dirty || gvctx->idlectx->dirty_saved) {
        strcat(msg, "start ");
    }
    if ((gvctx->idlectx->dirty  || gvctx->idlectx->dirty_saved) &&
        (gvctx->eidlectx->dirty || gvctx->eidlectx->dirty_saved))
    {
        strcat(msg, "and ");
        waswere = "times were ";
    }
    if (gvctx->eidlectx->dirty || gvctx->eidlectx->dirty_saved) {
        strcat(msg, "end ");
    }
    strcat(msg, waswere);
    strcat(msg, "modified        \n\n\n");
    strcat(msg, "   Record changes?\n");
    response = ___make_dialog_informational_run(gvctx, TRUE, msg);
    if (response == GTK_RESPONSE_OK) {
        gvctx->editctx.cursor = gvctx->idlectx->dirty_row ?
                                  gvctx->idlectx->dirty_row :
                                  gvctx->eidlectx->dirty_row;
        if (gvctx->idlectx->dirty_saved) {
            gvctx->idlectx->elapsed_sec  = gvctx->idlectx->dirty_sec;
            gvctx->idlectx->elapsed_usec = gvctx->idlectx->dirty_usec;
            gvctx->idlectx->ignore_adjustment_set = 1;
            cb_record_button(gvctx->idlectx->record, (gpointer) gvctx);
            gvctx->idlectx->ignore_adjustment_set = 0;
        }
        if (gvctx->eidlectx->dirty_saved) {
            gvctx->eidlectx->elapsed_sec  = gvctx->eidlectx->dirty_sec;
            gvctx->eidlectx->elapsed_usec = gvctx->eidlectx->dirty_usec;
            gvctx->eidlectx->ignore_adjustment_set = 1;
            cb_record_button(gvctx->eidlectx->record, (gpointer) gvctx);
            gvctx->eidlectx->ignore_adjustment_set = 0;
        }
    }
}


int ___row_modified_idleproc(gpointer data)
{
    global_vars_ctx *gvctx = (global_vars_ctx *) data;

#ifndef WIN32
    if (gdk_pointer_is_grabbed()) {
        /*
         * Do nothing until the cursor is ungrabbed.  This polling is
         * essential as this problem seems to be timing dependent.  Calling
         * this function only once usually works, but sometimes it is called
         * dozens of times. Remove this section at your peril.
         */
        return TRUE;
    }
#endif

    /*
     * Finally, call the modal dialog.  Return FALSE causes the idleproc to
     * quit.
     */
    ___row_modified_dialog_1(gvctx);

    /*
     * When provided, call function that is chained in to follow
     * complete processing of ___row_modified_dialog().
     */
    if (gvctx->row_modified_ctx.proc) {
        gvctx->row_modified_ctx.proc(gvctx->row_modified_ctx.data);
        gvctx->row_modified_ctx.proc = NULL;
    }
    return FALSE;
}


/*
 * This is a work-around for a well-known bug in GTK.  There is an unfortunate
 * interaction between GtkClist when GTK_SELECTION_BROWSE is enabled and a
 * modal dialog that causes the application to loose cursor focus.  This also
 * causes the X server to lock up.  This happens because both the clist and
 * the modal dialog are grabbing the cursor at the same time. The solution
 * implemented here is to start an idle proc to poll until the pointer is
 * ungrabbed.  This is a hack.
 *
 * This thread has the best discussion I've found yet for a work-around to
 * this problem.
 *
 *   http://mail.gnome.org/archives/gtk-list/2000-September/msg00283.html
 *   http://mail.gnome.org/archives/gtk-list/2000-September/msg00296.html
 *   http://mail.gnome.org/archives/gtk-list/2000-September/msg00298.html
 */
void ___row_modified_dialog(global_vars_ctx *gvctx)
{
    gtk_idle_add(___row_modified_idleproc, (gpointer) gvctx);
}


gint ___editor_dirty_dialog_run(global_vars_ctx *gvctx)
{
    gint response;

    response = ___make_dialog_informational_run(gvctx, TRUE,
                   "            Current session was changed            \n"
                   "           All modifications will be lost!         \n\n\n"
                   "  Discard changes?\n");
    return response;
                                                
}


void ___load_editor_blank_first_line(global_vars_ctx *gvctx)
{
    char          *dummy_argv[2];

    dummy_argv[0] = "";
    dummy_argv[1] = NULL;
    ___load_editor_by_files(gvctx, &dummy_argv[0], 1);
    gtk_clist_set_text(GTK_CLIST(gvctx->editctx.clist),
        gvctx->editctx.cursor->row_num, 0, "");
    gtk_clist_set_text(GTK_CLIST(gvctx->editctx.clist),
        gvctx->editctx.cursor->row_num, 1, "");
    gtk_clist_set_text(GTK_CLIST(gvctx->editctx.clist),
        gvctx->editctx.cursor->row_num, 2, "");
}


void cb_clear_editor(edit_menu_ctx *emctx)
{
    global_vars_ctx *gvctx;
    edit_row_ctx  *del;
    gint          response;
    edit_ctx      *editctx;

    if (!emctx) {
        return;
    }
    gvctx   = emctx->gvctx;
    editctx = &gvctx->editctx;

    cb_stop_button(gvctx->idlectx->stop_toggle, gvctx->idlectx);
    cb_stop_button(gvctx->eidlectx->stop_toggle, gvctx->eidlectx);
    
    if (editctx->dirty  ||
        gvctx->idlectx->dirty || gvctx->eidlectx->dirty)
    {
        response = ___editor_dirty_dialog_run(gvctx);
        if (response != GTK_RESPONSE_OK) {
            return;
        }
    }

    label_set_text(gvctx->idlectx->label_trt,
                   "Total Time: %d:%02d.%03d (%ds)", 0, 0, 0, 0);
    gtk_entry_set_text(GTK_ENTRY(gvctx->entry), "");
    gtk_tooltips_disable(gvctx->tooltips_entry);
        
    editctx->cursor = editctx->head;
    do {
        del = ___edit_row_list_delete(editctx);
    } while (del);
    gtk_clist_clear(GTK_CLIST(editctx->clist));
    editctx->dirty = FALSE;
    gtk_widget_set_sensitive(gvctx->menu_save_session, editctx->dirty);

    gtk_adjustment_set_value(GTK_ADJUSTMENT(gvctx->idlectx->adj), 0);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(gvctx->eidlectx->adj), 0);

    ___set_editor_defaults(gvctx);
    ___load_editor_blank_first_line(gvctx);
}


/*
 * This glue function is not strictly needed, but is used consistency
 * with the paradigm.
 */
void menuitem_response_rmd(void *data)
{
    void menuitem_response(edit_menu_ctx *emctx);

    menuitem_response((edit_menu_ctx *) data);
}


void menuitem_response(edit_menu_ctx *emctx)
{
    edit_row_ctx  *row;
    edit_row_ctx  *row_prev;
    int           insert;
    GtkAdjustment *adj;
    gdouble       incr;
    char          *item[3];
    global_vars_ctx *gvctx;
    int           action;
    edit_ctx      *editctx;
    edit_menu_ctx emctx_dup;

    if (!emctx) {
        return;
    }

    gvctx   = emctx->gvctx;
    action  = emctx->action;
    editctx = &gvctx->editctx;

#ifdef _DEBUG
    printf("menuitem_response %s\n", emctx->actionstr);
#endif
    if (action == EDIT_ACTION_DELETE) {
        if (!editctx->cursor) {
            return;
        }
        row = editctx->cursor;
        if (row->row_num == 0) {
            return;
        }
        if (editctx->cursor && 
            (gvctx->idlectx->dirty || gvctx->eidlectx->dirty))
        {
            gvctx->row_modified_ctx.proc = menuitem_response_rmd;
            gvctx->row_modified_ctx.data = emctx;
            ___row_modified_dialog(gvctx);
            return;
        }
        if (editctx->deleted) {
            ___row_free(editctx->deleted);
        }
        gtk_clist_remove(GTK_CLIST(editctx->clist), editctx->cursor->row_num);
        editctx->dirty = TRUE;
        editctx->deleted = ___edit_row_list_delete(editctx);
        ___edit_row_set_focus(gvctx);
        gtk_widget_set_sensitive(gvctx->menu_edit_pbelow, TRUE);
        gtk_widget_set_sensitive(gvctx->menu_save_session, editctx->dirty);
    }
    else if (action == EDIT_ACTION_PASTE || action == EDIT_ACTION_INSERT) {
        insert = (action == EDIT_ACTION_INSERT) ? 1 : 0;
        if (editctx->deleted) {
            if (gvctx->idlectx->dirty || gvctx->eidlectx->dirty) {
                gvctx->row_modified_ctx.proc = menuitem_response_rmd;
                gvctx->row_modified_ctx.data = emctx;
                ___row_modified_dialog(gvctx);
                return;
            }
            adj = gtk_scrolled_window_get_vadjustment(
                      GTK_SCROLLED_WINDOW(editctx->scrolled_window));
            row = editctx->deleted;
            editctx->deleted = NULL;
            ___edit_row_list_insert_cursor(row, editctx, insert);
            item[0] = row->file_name;
            item[1] = row->stime_str;
            item[2] = row->etime_str;
            gtk_clist_insert(GTK_CLIST(editctx->clist),
                             editctx->cursor->row_num, item);
            editctx->dirty = TRUE;
            ___edit_row_set_focus(gvctx);
            /*
             * Compute the amount the scroll window must be moved by to
             * see the entry after it is pasted in.  This is a bit
             * of a hack at this point.
             */
            incr = adj->upper / ((gdouble) editctx->tail->row_num) + 2.0;
            if (action == EDIT_ACTION_INSERT) {
                incr = -incr;
            }
            gtk_adjustment_set_value(adj, adj->value + incr);
            gtk_scrolled_window_set_vadjustment(
               GTK_SCROLLED_WINDOW(editctx->scrolled_window), adj);
            gtk_clist_select_row(GTK_CLIST(gvctx->editctx.clist),
                                 editctx->cursor->row_num, 0); 
            gtk_widget_set_sensitive(gvctx->menu_edit_pbelow, FALSE);
            gtk_widget_set_sensitive(gvctx->menu_save_session, editctx->dirty);
        }
    }
    else if (action == EDIT_ACTION_COPY || action == EDIT_ACTION_DUP) {
        row = editctx->cursor;
        if (row) {
            if (row->row_num == 0) {
                return;
            }
            if (gvctx->idlectx->dirty || gvctx->eidlectx->dirty) {
                gvctx->row_modified_ctx.proc = menuitem_response_rmd;
                gvctx->row_modified_ctx.data = emctx;
                ___row_modified_dialog(gvctx);
                return;
            }
            if (editctx->deleted) {
                ___row_free(editctx->deleted);
            }
            editctx->deleted = ___row_make(row->file_name, 0,
                                          row->s_min, row->s_sec, row->s_msec,
                                          row->e_min, row->e_sec, row->e_msec,
                                          G_CALLBACK(cb_row_clicked),
                                          (gpointer) gvctx);
#ifdef _DEBUG
            printf("after delete widget row_num is %d\n", 
                   editctx->cursor->row_num);
#endif
            if (action == EDIT_ACTION_DUP) {
                emctx_dup.action     = EDIT_ACTION_PASTE;
                emctx_dup.gvctx      = gvctx;
                emctx_dup.actionstr  = "Append";
                menuitem_response(&emctx_dup);
            }
            else {
                gtk_widget_set_sensitive(gvctx->menu_edit_pbelow, TRUE);
            }
        }
    }
    else if (action == EDIT_ACTION_TIME) {
        /*
         * Copy end time from previous entry to start time of current entry
         */
        if (editctx->cursor->row_num > 1) {
            row                          = editctx->cursor;
            row_prev                     = editctx->cursor->prev;
            gvctx->idlectx->elapsed_sec  = row_prev->e_min*60 +
                                               row_prev->e_sec;
            gvctx->idlectx->elapsed_usec = row_prev->e_msec;

            gvctx->idlectx->ignore_adjustment_set = 1;
            cb_record_button(gvctx->idlectx->record, (gpointer) gvctx);
            gvctx->idlectx->ignore_adjustment_set = 0;
            ___spin_scrollbar_set_value(gvctx->idlectx);
        }
    }
    else {
        ;
#ifdef _DEBUG
        printf("Unknown action %d\n", action);
#endif
    }
}


edit_menu_ctx *___make_emctx(global_vars_ctx *gvctx,
                             int action,
                             char *actionstr)
{
    edit_menu_ctx *emctx;

    emctx = calloc(1, sizeof(edit_menu_ctx));
    if (!emctx) {
        return NULL;
    }
    actionstr = strdup(actionstr);
    if (!actionstr) {
        free(emctx);
        return NULL;
    }
    emctx->gvctx     = gvctx;
    emctx->action    = action;
    emctx->actionstr = actionstr;
    return emctx;
}


/*
 * Generate a default edit output file name from the first entry in the
 * editor, when no output file has been set.
 */
void ___make_default_output_filename(global_vars_ctx *gvctx)
{
    char *name;
    char *new_name;
    char *cp;

    if (gvctx->output_file && !*gvctx->output_file && 
        gvctx->editctx.head && gvctx->editctx.head->next)
    {
        name = gvctx->editctx.head->next->file_name;
        cp   = strrchr(name, '/');
        if (cp) {
            name = cp + 1;
        }
        new_name = malloc(strlen(name) + 2);
        if (new_name) {
            strcpy(new_name, name);
            cp = strrchr(new_name, '.');
            if (cp) {
                strcpy(cp, "_");
            }
            if (gvctx->output_file) {
                free(gvctx->output_file);
            }
            gvctx->output_file = new_name;
        }
    }
}


void mainmenu_item_response(edit_menu_ctx *emctx)
{

#ifdef _DEBUG
    printf("mainmenu_item_response: action=%d\n", emctx->action);
#endif
    switch (emctx->action) {
      case MAINMENU_ACTION_OPEN:
        cb_open_file_from_menu(emctx->gvctx);
        break;

      case MAINMENU_ACTION_OUTFILE:
        ___make_default_output_filename(emctx->gvctx);
        gtk_entry_set_text(GTK_ENTRY(emctx->gvctx->ofdialogctx->entry),
                           emctx->gvctx->output_file);
        gtk_widget_show_all(emctx->gvctx->ofdialogctx->dialog);
        break;

      case MAINMENU_ACTION_LOAD_ABANDONED:
        gtk_widget_show_all(emctx->gvctx->abandon_file_ctx->dialog);
        break;

      case MAINMENU_ACTION_SAVE_ABANDON:
        emctx->gvctx->edit_action_abandon_save = TRUE;
        cb_edit_1(NULL, emctx->gvctx);
        break;

      case MAINMENU_ACTION_EDIT:
        cb_edit_1(NULL, emctx->gvctx);
        break;

      case MAINMENU_ACTION_CLEAR:
        cb_clear_editor(emctx);
        break;
      case MAINMENU_ACTION_ABOUT:
        gtk_widget_show_all(emctx->gvctx->about_dialog);
        gtk_dialog_run(GTK_DIALOG(emctx->gvctx->about_dialog));
        gtk_widget_hide_all(emctx->gvctx->about_dialog);
        break;
    }
}


void optionmenu_item_response(GtkWidget *widget, edit_menu_ctx *emctx)
{
    int state = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(widget));
    int state_change = 0;
    global_vars_ctx *gvctx = emctx->gvctx;

#ifdef _DEBUG
    printf("optionmenu_item_response: action=%d state=%d\n",
           emctx->action, state);
#endif
    switch (emctx->action) {
      case OPTIONMENU_ACTION_STOP:
        if (state != gvctx->stop_visable) {
            state_change = TRUE;
            gvctx->stop_visable = state;
        }
        if (state) {
            gtk_widget_show(gvctx->idlectx->stop_toggle);
            gtk_widget_show(gvctx->eidlectx->stop_toggle);
        }
        else {
            gtk_widget_hide(gvctx->idlectx->stop_toggle);
            gtk_widget_hide(gvctx->eidlectx->stop_toggle);
        }
        break;

      case OPTIONMENU_ACTION_PAUSE:
        if (state != gvctx->pause_visable) {
            state_change                = TRUE;
            gvctx->pause_visable = state;
        }
        if (state) {
            gtk_widget_show(gvctx->idlectx->pause_toggle);
            gtk_widget_show(gvctx->eidlectx->pause_toggle);
        }
        else {
            gtk_widget_hide(gvctx->idlectx->pause_toggle);
            gtk_widget_hide(gvctx->eidlectx->pause_toggle);
        }
        break;

      case OPTIONMENU_ACTION_VOLUME:
        if (state != gvctx->volume_visable) {
            state_change                = TRUE;
            gvctx->volume_visable = state;
        }
        if (state) {
            gtk_widget_show(gvctx->volumectx->box);
        }
        else {
            gtk_widget_hide(gvctx->volumectx->box);
        }
        break;
    }

    if (state_change) {
        write_config_file(CONFIG_FILE_NAME, gvctx);
    }
}



GtkWidget *___optionmenu_make(global_vars_ctx *gvctx)
{
    GtkWidget *menu;
    GtkWidget *menu_stop;
    GtkWidget *menu_pause;
    GtkWidget *menu_vol;
    edit_menu_ctx *omctx;

    menu         = gtk_menu_new();
    menu_stop    = gtk_check_menu_item_new_with_label("Stop button");
    menu_pause   = gtk_check_menu_item_new_with_label("Pause button");
    menu_vol     = gtk_check_menu_item_new_with_label("Volume Control");

    gtk_widget_show(menu_stop);
    gtk_widget_show(menu_pause);
    gtk_widget_show(menu_vol);

    gtk_menu_append(GTK_MENU(menu), menu_stop);
    gtk_menu_append(GTK_MENU(menu), menu_pause);
    gtk_menu_append(GTK_MENU(menu), menu_vol);

    omctx = ___make_emctx(gvctx, OPTIONMENU_ACTION_STOP, "Stop");
    g_signal_connect(G_OBJECT(menu_stop), "activate",
                     G_CALLBACK(optionmenu_item_response), omctx);
    omctx = ___make_emctx(gvctx, OPTIONMENU_ACTION_PAUSE, "Pause");
    g_signal_connect(G_OBJECT(menu_pause), "activate", 
                     G_CALLBACK(optionmenu_item_response), omctx);
    omctx = ___make_emctx(gvctx, OPTIONMENU_ACTION_VOLUME, "Volume");
    g_signal_connect(G_OBJECT(menu_vol), "activate", 
                     G_CALLBACK(optionmenu_item_response), omctx);

    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_stop),
                                   gvctx->stop_visable);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_pause),
                                   gvctx->pause_visable);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_vol),
                                   gvctx->volume_visable);
    return menu;
}


GtkWidget *___editmenu_make(global_vars_ctx *gvctx)
{
    GtkWidget     *menu;
    edit_menu_ctx *emctx;

    GtkWidget *menu_clear;
    GtkWidget *menu_edit_copy;
    GtkWidget *menu_edit_cut;
    GtkWidget *menu_edit_pbelow;
    GtkWidget *menu_edit_duplicate;
    GtkWidget *menu_edit_time;
    GtkWidget *menu_separator;
    GtkWidget *menu_separator2;

    menu             = gtk_menu_new();
    menu_edit_cut    = gtk_menu_item_new_with_label("Cut");
    menu_edit_copy   = gtk_menu_item_new_with_label("Copy");
    menu_edit_pbelow = gtk_menu_item_new_with_label("Append");
    menu_edit_duplicate = gtk_menu_item_new_with_label("Copy/Append");
    menu_separator   = gtk_separator_menu_item_new();
    menu_edit_time   = gtk_menu_item_new_with_label("Copy End Time");
    menu_separator2  = gtk_separator_menu_item_new();
    menu_clear       = gtk_menu_item_new_with_label("Clear editor");

    gtk_widget_show(menu_edit_cut);
    gtk_widget_show(menu_edit_copy);
    gtk_widget_show(menu_edit_pbelow);
    gtk_widget_show(menu_edit_duplicate);
    gtk_widget_show(menu_separator);
    gtk_widget_show(menu_edit_time);
    gtk_widget_show(menu_separator2);
    gtk_widget_show(menu_clear);

    gtk_menu_append(GTK_MENU(menu), menu_edit_cut);
    gtk_menu_append(GTK_MENU(menu), menu_edit_copy);
    gtk_menu_append(GTK_MENU(menu), menu_edit_pbelow);
    gtk_menu_append(GTK_MENU(menu), menu_edit_duplicate);
    gtk_menu_append(GTK_MENU(menu), menu_separator);
    gtk_menu_append(GTK_MENU(menu), menu_edit_time);
    gtk_menu_append(GTK_MENU(menu), menu_separator2);
    gtk_menu_append(GTK_MENU(menu), menu_clear);

    emctx = ___make_emctx(gvctx, EDIT_ACTION_DELETE, "Delete");
    g_signal_connect_swapped(G_OBJECT(menu_edit_cut), "activate", 
                             G_CALLBACK(menuitem_response), emctx);
    emctx = ___make_emctx(gvctx, EDIT_ACTION_COPY, "Copy");
    g_signal_connect_swapped(G_OBJECT(menu_edit_copy), "activate", 
                             G_CALLBACK(menuitem_response), emctx);
    emctx = ___make_emctx(gvctx, EDIT_ACTION_PASTE, "Append");
    g_signal_connect_swapped(G_OBJECT(menu_edit_pbelow), "activate", 
                             G_CALLBACK(menuitem_response), emctx);
    emctx = ___make_emctx(gvctx, EDIT_ACTION_DUP, "Copy/Append");
    g_signal_connect_swapped(G_OBJECT(menu_edit_duplicate), "activate", 
                             G_CALLBACK(menuitem_response), emctx);
    emctx = ___make_emctx(gvctx, EDIT_ACTION_TIME, "Copy End Time");
    g_signal_connect_swapped(G_OBJECT(menu_edit_time), "activate", 
                             G_CALLBACK(menuitem_response), emctx);
    emctx = ___make_emctx(gvctx, MAINMENU_ACTION_CLEAR, "Clear editor");
    g_signal_connect_swapped(G_OBJECT(menu_clear), "activate", 
                             G_CALLBACK(mainmenu_item_response), emctx);

    gtk_widget_set_sensitive(menu_edit_pbelow,    FALSE);
    gtk_widget_set_sensitive(menu_edit_cut,       FALSE);
    gtk_widget_set_sensitive(menu_edit_copy,      FALSE);
    gtk_widget_set_sensitive(menu_edit_duplicate, FALSE);
    gtk_widget_set_sensitive(menu_edit_time,      FALSE);

    gvctx->menu_edit_pbelow    = menu_edit_pbelow;
    gvctx->menu_edit_cut       = menu_edit_cut;
    gvctx->menu_edit_copy      = menu_edit_copy;
    gvctx->menu_edit_duplicate = menu_edit_duplicate;
    gvctx->menu_edit_time      = menu_edit_time;
    return menu;
}


GtkWidget *___helpmenu_make(global_vars_ctx *gvctx)
{
    GtkWidget *menu;
    GtkWidget *menu_about;
    edit_menu_ctx *mmctx;

    menu           = gtk_menu_new();
    menu_about     = gtk_menu_item_new_with_label("About");

    gtk_widget_show(menu_about);

    gtk_menu_append(GTK_MENU(menu), menu_about);

/* probably wrong callback action for this menu */
    mmctx = ___make_emctx(gvctx, MAINMENU_ACTION_ABOUT, "About");
    g_signal_connect_swapped(G_OBJECT(menu_about), "activate", 
                             G_CALLBACK(mainmenu_item_response), mmctx);
    return menu;
}


GtkWidget *___mainmenu_make(global_vars_ctx *gvctx)
{
    GtkWidget *menu;
    GtkWidget *menu_open;
    GtkWidget *menu_edit;
    GtkWidget *menu_separator1;
    GtkWidget *menu_separator2;
    GtkWidget *menu_separator3;
    GtkWidget *menu_separator4;
    GtkWidget *menu_separator5;
    GtkWidget *menu_options;
    GtkWidget *menu_outfile;
    GtkWidget *menu_exit;
    GtkWidget *menu_load_abandoned;
    GtkWidget *option_menu;
    edit_menu_ctx *mmctx;

    menu            = gtk_menu_new();
    menu_open       = gtk_menu_item_new_with_label("Open File...");
    menu_separator5 = gtk_separator_menu_item_new();
    menu_outfile    = gtk_menu_item_new_with_label("Output file...");
    menu_separator1 = gtk_separator_menu_item_new();

    gvctx->menu_save_session = 
                     gtk_menu_item_new_with_label("Save session...");
    menu_load_abandoned =
                     gtk_menu_item_new_with_label("Load session...");
    menu_separator2 = gtk_separator_menu_item_new();

    menu_edit       = gtk_menu_item_new_with_label("Edit");
    menu_separator3 = gtk_separator_menu_item_new();

    menu_options    = gtk_menu_item_new_with_label("Options");
    menu_separator4 = gtk_separator_menu_item_new();

    menu_exit       = gtk_menu_item_new_with_label("Exit");

    gtk_widget_show(menu_open);
    gtk_widget_show(menu_separator5);
    gtk_widget_show(menu_outfile);
    gtk_widget_show(menu_separator1);
    gtk_widget_show(gvctx->menu_save_session);
    gtk_widget_show(menu_load_abandoned);
    gtk_widget_show(menu_separator2);
    gtk_widget_show(menu_edit);
    gtk_widget_show(menu_separator3);
    gtk_widget_show(menu_options);
    gtk_widget_show(menu_separator4);
    gtk_widget_show(menu_exit);

    gtk_menu_append(GTK_MENU(menu), menu_open);
    gtk_menu_append(GTK_MENU(menu), menu_separator5);
    gtk_menu_append(GTK_MENU(menu), menu_outfile);
    gtk_menu_append(GTK_MENU(menu), menu_separator1);
    gtk_menu_append(GTK_MENU(menu), gvctx->menu_save_session);
    gtk_menu_append(GTK_MENU(menu), menu_load_abandoned);
    gtk_menu_append(GTK_MENU(menu), menu_separator2);
    gtk_menu_append(GTK_MENU(menu), menu_edit);
    gtk_menu_append(GTK_MENU(menu), menu_separator3);
    gtk_menu_append(GTK_MENU(menu), menu_options);
    gtk_menu_append(GTK_MENU(menu), menu_separator4);
    gtk_menu_append(GTK_MENU(menu), menu_exit);

    gtk_widget_set_sensitive(gvctx->menu_save_session, FALSE);

    mmctx = ___make_emctx(gvctx, MAINMENU_ACTION_OPEN,
                          "Open file");
    g_signal_connect_swapped(G_OBJECT(menu_open), "activate",
                             G_CALLBACK(mainmenu_item_response), mmctx);
    mmctx = ___make_emctx(gvctx, MAINMENU_ACTION_LOAD_ABANDONED,
                          "Load session");
    g_signal_connect_swapped(G_OBJECT(menu_load_abandoned), "activate", 
                             G_CALLBACK(mainmenu_item_response), mmctx);
    mmctx = ___make_emctx(gvctx, MAINMENU_ACTION_SAVE_ABANDON,
                          "Save edit session");
    g_signal_connect_swapped(G_OBJECT(gvctx->menu_save_session), "activate", 
                             G_CALLBACK(mainmenu_item_response), mmctx);
    mmctx = ___make_emctx(gvctx, MAINMENU_ACTION_OUTFILE, "Output file");
    g_signal_connect_swapped(G_OBJECT(menu_outfile), "activate", 
                             G_CALLBACK(mainmenu_item_response), mmctx);

    option_menu = ___optionmenu_make(gvctx);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_options),
                              option_menu);
    mmctx = ___make_emctx(gvctx, MAINMENU_ACTION_EDIT, "Edit");
    g_signal_connect_swapped(G_OBJECT(menu_edit), "activate", 
                             G_CALLBACK(mainmenu_item_response), mmctx);

    g_signal_connect(G_OBJECT(menu_exit), "activate",
                     G_CALLBACK(cb_quit), gvctx->idlectx);

    gvctx->menu_edit = menu_edit;

    gtk_widget_set_sensitive(menu_edit, FALSE);
    return menu;
}


GtkWidget *___aboutdialog_make(global_vars_ctx *gvctx)
{
    GtkWidget *dialog;
    GtkWidget *label1;
    GtkWidget *label2;
    GtkWidget *scrolled_window;


    dialog = gtk_dialog_new_with_buttons(
                 "About",
                 GTK_WINDOW(gvctx->main_window), GTK_DIALOG_MODAL,
                 GTK_STOCK_OK,     GTK_RESPONSE_OK,
                 NULL);
    scrolled_window = gtk_scrolled_window_new(NULL, NULL);
    label1 = gtk_label_new( 
      "                     " XMPGEDIT_PRODUCT "\n\n"
                            MPGEDIT_COPYRIGHT);
    label2 = gtk_label_new(
      "Product:         " XMPGEDIT_PRODUCT "\n"
      "Version:         " XMPGEDIT_VERSION " " MPGEDIT_URL"\n"
      "Build:            " MPGEDIT_BUILDNUM "("MPGEDIT_BUILDOS")\n"
      "Report bugs to:  " MPGEDIT_MAILTO "\n\n"
      XMPGEDIT_CREDITS_COPYRIGHT);

    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
                       label1, FALSE, FALSE, 5);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
                       scrolled_window, FALSE, FALSE, 5);

    gtk_scrolled_window_add_with_viewport(
                   GTK_SCROLLED_WINDOW(scrolled_window), label2);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
                                   GTK_POLICY_NEVER,
                                   GTK_POLICY_AUTOMATIC);
    gtk_widget_set_size_request(scrolled_window, -1, 200);
    return dialog;
}



/*
 * Enforce exclusive use of the play/playto buttons.  When start control
 * is playing, prevent playback from the end control, and vice versa.
 */
int ___button_used_match(GtkWidget *widget,
                         GtkWidget *match,
                         idle_ctx *idlectx)
{
    if (match) {
        if (match != widget) {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), FALSE);
            return 0;
        }
    }
    else {
        idlectx->gvctx->playto_toggle = idlectx->playto_toggle;
        idlectx->gvctx->play_toggle   = idlectx->play_toggle;
        idlectx->gvctx->pause_toggle  = idlectx->pause_toggle;
        idlectx->gvctx->stop_toggle   = idlectx->stop_toggle;
    }

    return 1;
}


/*
 * Clear exclusive use state from play/playto buttons.
 */
void ___button_used_clear(idle_ctx *idlectx)
{
    idlectx->gvctx->playto_toggle = NULL;
    idlectx->gvctx->play_toggle   = NULL;
    idlectx->gvctx->pause_toggle  = NULL;
    idlectx->gvctx->stop_toggle   = NULL;
}


void ___spin_button_set_value(idle_ctx *idlectx, int insec, int inmsec)
{
    long min  = insec / 60;
    long sec  = insec % 60;
    long msec = inmsec;

    idlectx->ignore_spin_button_callbacks = 1;
    gtk_spin_button_set_value(
        GTK_SPIN_BUTTON(idlectx->sbctx.minute), min);

    gtk_spin_button_set_value(
        GTK_SPIN_BUTTON(idlectx->sbctx.second), sec);

    gtk_spin_button_set_value(
        GTK_SPIN_BUTTON(idlectx->sbctx.millisecond), msec);
    idlectx->ignore_spin_button_callbacks = 0;
    
    idlectx->sbctx.minute_prev      = min;
    idlectx->sbctx.second_prev      = sec;
    idlectx->sbctx.millisecond_prev = msec;
}


void ___spin_scrollbar_set_value(idle_ctx *idlectx)
{
    gdouble t_value;
    gdouble t_min;
    gdouble t_sec;
    gdouble t_msec;

    t_value  = t_min = gtk_spin_button_get_value(
                             GTK_SPIN_BUTTON(idlectx->sbctx.minute))* 60.0;
    t_value += t_sec = gtk_spin_button_get_value(
                             GTK_SPIN_BUTTON(idlectx->sbctx.second));
    t_msec = gtk_spin_button_get_value(
                 GTK_SPIN_BUTTON(idlectx->sbctx.millisecond));
    t_value += t_msec / 1000.0;
    gtk_adjustment_set_value(GTK_ADJUSTMENT(idlectx->adj), t_value);
    gtk_adjustment_changed(GTK_ADJUSTMENT(idlectx->adj));
}




void label_set_text(GtkWidget *widget, char *msg, ...)
{
    va_list ap;
    char buf[128];

    va_start(ap, msg);
    vsprintf(buf, msg, ap);
    gtk_label_set_text(GTK_LABEL(widget), buf);
    va_end(ap);
}



int cb_label_set_text(void *data, long sec, long usec)
{
    idle_ctx *idlectx = (idle_ctx *) data;

    idlectx->elapsed_sec  = sec;
    idlectx->elapsed_usec = usec;
    if (idlectx->saved_sec != sec) {
        idlectx->saved_sec = sec;
    
        label_set_text(idlectx->label_elapsed, "%d:%02d.%03d (%ds)",
                       sec/60, sec%60, usec, sec);
    }
    return 1;
}


gint idle_play(gpointer data)
{
    idle_ctx *idlectx = (idle_ctx *) data;
    long sec;
    long msec;
    gint go = TRUE;

    if (mpgedit_play_frame(idlectx->playctx) == 0) {
        go   = FALSE;
        sec  = mpgedit_play_current_sec(idlectx->playctx);
        msec = mpgedit_play_current_msec(idlectx->playctx);
        /*
         * This is a hack, but needed due to how playback is stopped
         * when the time is modified by cb_scrollbar_adj().  There should
         * be a better way to stop playback, other than just calling
         * mpgedit_play_close().  Anyway, it is kind of difficult to get
         * a non-zero sec/msec value after play_close has been called.
         */
        if (sec) {
            idlectx->elapsed_sec  = sec;
        }
        if (msec) {
            idlectx->elapsed_usec = msec;
        }
        mpgedit_play_close(idlectx->playctx);
        idlectx->playctx = NULL; 
        idlectx->play_toggle_ignore = 1;
        gtk_toggle_button_set_active(
            GTK_TOGGLE_BUTTON(idlectx->play_toggle), FALSE);
        idlectx->play_toggle_ignore = 0;
        ___button_used_clear(idlectx);
    }
    return go;
}


gint idle_playto(gpointer data)
{
    idle_ctx *idlectx = (idle_ctx *) data;
    long sec;
    long msec;
    

    sec  = mpgedit_play_current_sec(idlectx->playctx);
    msec = mpgedit_play_current_msec(idlectx->playctx);

    if ((sec  >= idlectx->stime_sec &&
         msec >= idlectx->stime_usec)
        || (mpgedit_play_frame(idlectx->playctx) == 0))
    {
        idlectx->elapsed_sec  = idlectx->stime_sec;
        idlectx->elapsed_usec = idlectx->stime_usec;
        mpgedit_play_close(idlectx->playctx);
        idlectx->playctx = NULL; 
        gtk_idle_remove(idlectx->idle_tag);
        gtk_toggle_button_set_active(
            GTK_TOGGLE_BUTTON(idlectx->playto_toggle), FALSE);
        ___button_used_clear(idlectx);
    }
    return TRUE;
}


void cb_record_button(GtkWidget *widget, gpointer *data)
{
/* adam/TBD: "data" should be idlectx or eidlectx, not gvctx. This will
 * eliminate the need to "decode" which context is needing changing.
 */
    global_vars_ctx *gvctx = (global_vars_ctx *) data;
    idle_ctx        *idlectx;
    long            sec;
    long            usec;
    int             changed;

    /* 
     * Figure out which playback control we are; start or end playback.
     */
    idlectx = (gvctx->idlectx->record == widget) ? 
        gvctx->idlectx : gvctx->eidlectx;

    idlectx->stime_sec  = sec  = idlectx->elapsed_sec;
    idlectx->stime_usec = usec = idlectx->elapsed_usec;

    label_set_text(idlectx->label_start, "%d:%02d.%03d (%ds)",
                   sec/60, sec%60, usec, sec);

    /*
     * Must manipulate playback controls separately.  Both are available
     * in the gvctx. The first is the start time control, the other is
     * the end time control.
     */
    if (gvctx->idlectx->record == widget) {
/*
 * This sucks!! Need reference to the portion of the row labels in the 
 * idlectx that applies to the playback control.
 */
        changed = ___row_init(gvctx->editctx.cursor, NULL, 
                              sec/60, sec%60, usec,
                              -1, -1, -1);
    }
    else {
        changed = ___row_init(gvctx->editctx.cursor, NULL, 
                              -1, -1, -1,
                              sec/60, sec%60, usec);
    }

    idlectx->dirty = idlectx->dirty_saved = FALSE;
    /*
     * Nothing has changed, so just quit this callback
     */
    if (!changed) {
        return;
    }
    ___spin_button_set_value(idlectx,
                             idlectx->stime_sec,
                             idlectx->stime_usec);

    ___row_update(&gvctx->editctx, gvctx->editctx.clist);
    gvctx->editctx.dirty = TRUE;
    gtk_widget_set_sensitive(gvctx->menu_save_session, gvctx->editctx.dirty);
}


/*
 * Play from 5 seconds before current start time to the currently selected
 * start time.  This callback is similar to cb_play_button(), but uses a
 * different start time criteria, and uses a different idle function,
 * idle_playto(), to do the playback.  idle_playto() stops when the play
 * runtime exceeds the start time.
 */
void cb_playto_button(GtkWidget *widget, gpointer *data)
{
    idle_ctx *idlectx = (idle_ctx *) data;
    long sec;
    long usec;

    int playing;

    if (!___button_used_match(widget, idlectx->gvctx->playto_toggle, idlectx)) {
        return;
    }

    /*
     * Quit if play is running
     */
    playing = gtk_toggle_button_get_active(
                  GTK_TOGGLE_BUTTON(idlectx->play_toggle));
    if (playing) {
        gtk_toggle_button_set_active(
            GTK_TOGGLE_BUTTON(idlectx->playto_toggle), FALSE);
        return;
    }

    playing = gtk_toggle_button_get_active(
                  GTK_TOGGLE_BUTTON(idlectx->playto_toggle));
    if (playing) {

        /*
         * Transitioned from stopped -> playing
         */
        if (!idlectx->play_file || 
            (idlectx->stime_sec == 0 && idlectx->stime_usec == 0))
        {
            gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(idlectx->playto_toggle), FALSE);
            ___button_used_clear(idlectx);
            return;
        }

        /*
         * Seek to desired play start time. For this button, that is
         * start playing 5 seconds before the current start time.
         */
        sec  = idlectx->stime_sec - 5;
        usec = idlectx->stime_usec;
        if (sec < 0) {
            sec  = 0;
            usec = 0;
        }

        idlectx->playctx = mpgedit_play_init(idlectx->play_file);
        if (!idlectx->playctx) {
            ___make_cant_open_sound(idlectx->gvctx);
            /*
             * Do something more exciting when this fails
             */
            gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(idlectx->playto_toggle), FALSE);
            ___button_used_clear(idlectx);
            return;
        }

        mpgedit_play_seek_time(idlectx->playctx, sec, usec);

#ifdef _DEBUG
        printf("cb_playto_button: size=%ld sec=%ld usec=%ld\n", 
               mpgedit_play_total_size(idlectx->playctx),
               mpgedit_play_total_sec(idlectx->playctx),
               mpgedit_play_total_msec(idlectx->playctx));
#endif

        idlectx->idle_tag = gtk_idle_add(idle_playto, idlectx);
        mpgedit_play_set_status_callback(idlectx->playctx,
                                         cb_label_set_text, 
                                         idlectx);
    }
    else {
        /*
         * Transitioned from playing -> stopped
         */
        idlectx->elapsed_sec  = idlectx->stime_sec;
        idlectx->elapsed_usec = idlectx->stime_usec;
        label_set_text(idlectx->label_elapsed, "%d:%02d.%03d (%ds)", 
                       idlectx->stime_sec/60, 
                       idlectx->stime_sec%60, 
                       idlectx->stime_usec,
                       idlectx->stime_sec);
        
        mpgedit_play_close(idlectx->playctx);
        idlectx->playctx = NULL; 
        if (idlectx->idle_tag) {
            gtk_idle_remove(idlectx->idle_tag);
        }
        gtk_toggle_button_set_active(
            GTK_TOGGLE_BUTTON(idlectx->pause_toggle), FALSE);

        ___button_used_clear(idlectx);
    }
    return;
}


void cb_spin_modified(GtkWidget *widget, gpointer *data)
{
    idle_ctx *idlectx = (idle_ctx *) data;

    if (idlectx->ignore_adjustment_set) {
        return;
    }
    idlectx->dirty       = TRUE;
    idlectx->dirty_saved = TRUE;
    idlectx->dirty_sec   = idlectx->elapsed_sec;
    idlectx->dirty_usec  = idlectx->elapsed_usec;
    idlectx->dirty_row   = idlectx->gvctx->editctx.cursor;
}


void cb_spin_minute(GtkWidget *widget, gpointer *data)
{
    idle_ctx *idlectx = (idle_ctx *) data;
    gdouble  minute;
    int      delta;

    if (idlectx->ignore_spin_button_callbacks) {
        return;
    }

    minute = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
    delta  = (int) (minute - idlectx->sbctx.minute_prev);
    if (idlectx->sbctx.minute_prev + delta <= 0.0) {
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), 0.0);
        idlectx->sbctx.minute_prev = 0.0;
        return;
    }

    idlectx->sbctx.minute_prev = minute;
    idlectx->stime_sec        += delta * 60;
    idlectx->ignore_spin_button_set = 1;
    ___spin_scrollbar_set_value(idlectx);
    idlectx->ignore_spin_button_set = 0;
}


void cb_spin_second(GtkWidget *widget, gpointer *data)
{
    idle_ctx *idlectx = (idle_ctx *) data;
    gdouble  second;
    gdouble  delta;

    if (idlectx->ignore_spin_button_callbacks) {
        return;
    }

    second = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
    delta  = second - idlectx->sbctx.second_prev;
    if (second <= 0.0 && idlectx->sbctx.minute_prev <= 0.0) {
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), 0.0);
        idlectx->sbctx.second_prev = 0.0;
    }
    else if (second >= 60.0) {
        gtk_spin_button_spin(GTK_SPIN_BUTTON(idlectx->sbctx.minute),
                             GTK_SPIN_STEP_FORWARD, 1.0);
        idlectx->sbctx.second_prev = 0.0;
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), 0.0);
    }
    else if ((second < idlectx->sbctx.second_prev) &&
              idlectx->sbctx.millisecond_prev > 0)
    {
        /*
         * When decrementing seconds, zero millisecond field when its
         * value is not already zero.  When seconds is decremented again,
         * then seconds value will change.
         */
        gtk_spin_button_set_value(
            GTK_SPIN_BUTTON(idlectx->sbctx.millisecond), 0.0);
        idlectx->sbctx.millisecond_prev = 0;
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(idlectx->sbctx.second),
                                  idlectx->sbctx.second_prev);
        
    }
    else if (second < 0.0) {
        gtk_spin_button_spin(GTK_SPIN_BUTTON(idlectx->sbctx.minute),
                             GTK_SPIN_STEP_BACKWARD, 1.0);
        second = 60.0 + delta;
        idlectx->sbctx.second_prev = second;
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), second);
    }
    else {
        idlectx->sbctx.second_prev = second;
    }
    /*
     * Wipe out any millisecond value when second value has changed
     */
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(idlectx->sbctx.millisecond), 0.0);

    idlectx->stime_sec += (int) delta;
    idlectx->ignore_spin_button_set = 1;
    ___spin_scrollbar_set_value(idlectx);
    idlectx->ignore_spin_button_set = 0;
}


void cb_spin_millisecond(GtkWidget *widget, gpointer *data)
{
    idle_ctx *idlectx = (idle_ctx *) data;
    gdouble  millisecond;
    gdouble  delta;

    if (idlectx->ignore_spin_button_callbacks) {
        return;
    }

    millisecond = gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
    delta       = millisecond - idlectx->sbctx.millisecond_prev;
    if (idlectx->sbctx.millisecond_prev + delta <= 0.0 &&
        idlectx->sbctx.second_prev <= 0.0 &&
        idlectx->sbctx.minute_prev <= 0.0)
    {
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), 0.0);
        idlectx->sbctx.millisecond_prev = 0.0;
    }
    else if (millisecond >= 1000.0) {
        gtk_spin_button_spin(GTK_SPIN_BUTTON(idlectx->sbctx.second),
                             GTK_SPIN_STEP_FORWARD, 1.0);
        idlectx->sbctx.millisecond_prev = 0.0;
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), 0.0);
    }
    else if (millisecond < 0.0) {
        gtk_spin_button_spin(GTK_SPIN_BUTTON(idlectx->sbctx.second),
                             GTK_SPIN_STEP_BACKWARD, 1.0);
        millisecond = 1000.0 + delta;
        idlectx->sbctx.millisecond_prev = millisecond;
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), millisecond);
    }
    else {
        idlectx->sbctx.millisecond_prev = millisecond;
    }
    idlectx->stime_usec += (int) delta;
    idlectx->ignore_spin_button_set = 1;
    ___spin_scrollbar_set_value(idlectx);
    idlectx->ignore_spin_button_set = 0;
}



void cb_scrollbar_adj(GtkAdjustment *adj, gpointer *data)
{
    idle_ctx *idlectx = (idle_ctx *) data;
    long     sec      = (long) adj->value;
    long     msec     = (long) ((adj->value - ((gdouble) sec)) * 1000.0);
    int      playing;

    playing  = gtk_toggle_button_get_active(
                   GTK_TOGGLE_BUTTON(idlectx->play_toggle));
    if (playing) {
        /*
         * adam: Stopping playback by calling mpgedit_play_close() here is
         * somewhat bogus.  Adding a play flag to the idlectx that is checked
         * in idle() is a better solution.  All of the calls to gtk_idle_remove()
         * should be examined in the future.
         */
        mpgedit_play_close(idlectx->playctx);
        gtk_toggle_button_set_active(
            GTK_TOGGLE_BUTTON(idlectx->pause_toggle), FALSE);
        idlectx->playctx = NULL; 
    }
    label_set_text(idlectx->label_start,   "%d:%02d.%03d (%ds)", 
                   sec/60, sec%60, msec, sec);
    if (!idlectx->adj_elapsed_label_noset) {
        label_set_text(idlectx->label_elapsed, "%d:%02d.%03d (%ds)", 
                       sec/60, sec%60, msec, sec);
    }
    idlectx->stime_sec    = sec;
    idlectx->elapsed_sec  = sec;
    idlectx->stime_usec   = msec;
    idlectx->elapsed_usec = msec;
    if (!idlectx->ignore_spin_button_set) {
        ___spin_button_set_value(idlectx, sec, msec);
    }
}


/*
 * Convenience function to initialize the player contexts when the 
 * file selection widget OK button is pushed.
 * Inputs:
 *   idlectx->play_file
 */
void ___file_sel_ok(GtkWidget *widget, gpointer *data,
                    long init_sec, long init_msec)
{
    idle_ctx *idlectx = (idle_ctx *) data;
    void     *playctx = NULL;
    long     sec;
    long     usec;
    int      playing;

    if (___button_used_match(idlectx->play_toggle, 
                             idlectx->gvctx->play_toggle, idlectx))
    {
        ___button_used_clear(idlectx);
        /*
         * Stop playing file if playing
         */
        playing  = gtk_toggle_button_get_active(
                       GTK_TOGGLE_BUTTON(idlectx->play_toggle));
        if (playing) {
            gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(idlectx->pause_toggle), FALSE);
            gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(idlectx->play_toggle), FALSE);
            idlectx->playctx = NULL; 
        }
        playing  = gtk_toggle_button_get_active(
                       GTK_TOGGLE_BUTTON(idlectx->playto_toggle));
        if (playing) {
            gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(idlectx->pause_toggle), FALSE);
            gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(idlectx->playto_toggle), FALSE);
            idlectx->playctx = NULL; 
        }
    
        /*
         * Set the playback scroll bar range using the input file play length
         */
        playctx = mpgedit_play_init(idlectx->play_file);
        if (!playctx) {
            return;
        }
        sec  = mpgedit_play_total_sec(playctx);
        usec = mpgedit_play_total_msec(playctx);
        mpgedit_play_close(playctx);
        idlectx->playctx = NULL;
        if (!idlectx->adj_elapsed_label_noset) {
            label_set_text(idlectx->label_trt,
                           "Total Time: %d:%02d.%03d (%ds)",
                           sec/60, sec%60, usec, sec);
        }

        GTK_ADJUSTMENT(idlectx->adj)->upper = (float) sec + 1.0;
    }
    idlectx->ignore_adjustment_set = 1;
    if (init_sec || init_msec) {
        ___spin_button_set_value(idlectx, init_sec, init_msec);
        idlectx->ignore_spin_button_set = 1;
        ___spin_scrollbar_set_value(idlectx);
        idlectx->ignore_spin_button_set = 0;
    }
    else {
        gtk_adjustment_set_value(GTK_ADJUSTMENT(idlectx->adj), 0.0);
    }
    idlectx->ignore_adjustment_set = 0;

    gtk_toggle_button_set_active(
        GTK_TOGGLE_BUTTON(idlectx->gvctx->open_file_toggle), FALSE);

    if (!idlectx->adj_elapsed_label_noset) {
        label_set_text(idlectx->label_elapsed, "%d:%02d.%03d (%ds)", 
                       idlectx->elapsed_sec/60,
                       idlectx->elapsed_sec%60,
                       idlectx->elapsed_usec,
                       idlectx->elapsed_sec);

    }
}


void cb_edit_menu(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    GdkEventButton  *event_button = (GdkEventButton *) event;
    global_vars_ctx *gvctx        = (global_vars_ctx *) data;

#ifdef _DEBUG
    printf("cb_edit_menu: event type: %d\n", event->type);
#endif
    if (event->type == GDK_BUTTON_PRESS) {
#ifdef _DEBUG
        printf("cb_edit_menu: event_button: %d\n", event_button->button);
#endif
        if (event_button->button == 3) {
            gtk_menu_popup (GTK_MENU (gvctx->edit_menu), NULL, NULL, NULL,
                            NULL, event_button->button, event_button->time);

        }
    }
}


void cb_row_clicked_1_rmd(gpointer data)
{
    void cb_row_clicked_1a(GtkWidget *, GdkEvent *, gpointer);
    edit_row_ctx    *er      = (edit_row_ctx *)    data;
    global_vars_ctx *gvctx   = (global_vars_ctx *) er->ctx;

    cb_row_clicked_1a(gvctx->row_modified_ctx.widget,
                      gvctx->row_modified_ctx.event,
                      data);
}


void cb_row_clicked_1a(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    edit_row_ctx    *er      = (edit_row_ctx *) data;
    global_vars_ctx *gvctx   = (global_vars_ctx *) er->ctx;

    gvctx->idlectx->dirty  = FALSE;
    gvctx->eidlectx->dirty = FALSE;
    gvctx->editctx.cursor  = er;
    ___edit_row_set_focus(gvctx);
}


void cb_row_clicked_1(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    edit_row_ctx    *er      = (edit_row_ctx *) data;
    global_vars_ctx *gvctx   = (global_vars_ctx *) er->ctx;

    if (gvctx->editctx.cursor &&
        (gvctx->idlectx->dirty || gvctx->eidlectx->dirty))
    {
#ifdef _DEBUG
        printf("cb_row_clicked_1: stime=%d:%d s=%d:%d.%d e=%d:%d.%d\n",
               gvctx->idlectx->stime_sec, 
               gvctx->idlectx->stime_usec, 
               er->s_min, er->s_sec, er->s_msec, 
               er->e_min, er->e_sec, er->e_msec);
#endif
        gvctx->row_modified_ctx.proc   = cb_row_clicked_1_rmd;
        gvctx->row_modified_ctx.data   = data;
        gvctx->row_modified_ctx.widget = widget;
        gvctx->row_modified_ctx.event  = event;
        ___row_modified_dialog(gvctx);
        return;
    }

    cb_row_clicked_1a(widget, event, data);
}


idle_indexfile_ctx *___index_file_list_get(global_vars_ctx *gvctx)
{
    index_file_list *node;
    idle_indexfile_ctx *ifctx;

    if (!gvctx || !gvctx->ifl_head) {
        return NULL;
    }

    node            = gvctx->ifl_head;
    gvctx->ifl_head = gvctx->ifl_head->next;
    if (!gvctx->ifl_head) {
        gvctx->ifl_tail = NULL;
    }
    
    ifctx = node->ifctx;
    free(node);

    return ifctx;
}


int ___index_file_list_put(global_vars_ctx *gvctx, idle_indexfile_ctx *ifctx)
{
    index_file_list *node;

    if (!gvctx || !ifctx) {
        return 1;
    }

    node = (index_file_list *) calloc(1, sizeof(index_file_list));
    if (!node) {
        return 1;
    }
    node->ifctx = ifctx;
    if (!gvctx->ifl_head) {
        gvctx->ifl_head = node;
        gvctx->ifl_tail = node;
    }
    else {
        gvctx->ifl_tail->next = node;
        gvctx->ifl_tail       = node;
    }
    return 0;
}


/*
 * Idle proc that performs mp3 file indexing.  This is initialized by
 * the cb_row_clicked() callback.  Once this indexing operation is
 * complete, the remainder of the callback function, cb_row_clicked_1()
 * is called.
 */
gint row_clicked_idle_indexfile(gpointer data)
{
    indexfile_idleproc_ctx *ipctx = (indexfile_idleproc_ctx *) data;
    global_vars_ctx        *gvctx = ipctx->gvctx;
    idle_indexfile_ctx     *ctx;
    long                   tnow;
    long                   frames;
    gdouble                percent;
    long                   offset;
    int                    eof;
    char                   *label;
    struct stat            sb;

    if (!ipctx->ifctx) {
        ctx          = ___index_file_list_get(gvctx);
        ipctx->ifctx = ctx;
    }
    else {
        ctx = ipctx->ifctx;
    }

    if (!ctx) {
        /* Delete this idle proc when nothing left to process */
        if (ipctx->dialog) {
            gtk_widget_destroy(ipctx->dialog);
        }
        free(ipctx);
        gvctx->idle_indexfile_count--;
        return 0;
    }

    /*
     * This condition is true when cb_edit_1a() creates a "fake" edit_row
     * context.  The important value here is the callback.
     */
    if (!ctx->file_name && ctx->callback) {
        if (ctx->callback) {
            ctx->callback(ctx->widget, ctx->event, ctx->data);
        }
        free(ctx);
        ipctx->ifctx = NULL;
        return 1;
    }

    if (!ctx->playctx) {
        if (stat(ctx->file_name, &sb) != -1) {
            ctx->fsize = sb.st_size;
        }
        ctx->playctx = mpgedit_edit_index_init(
                           (char *) ctx->edit_row->file_name);
        eof = mpgedit_edit_index(ctx->playctx);
        if (!eof && !ipctx->dialog) {
            /*
             * Create dialog box with progress bar
             */
            ipctx->dialog = ___make_indexfile_progress_dialog(
                                ipctx,
                                gvctx->main_window,
                                ctx->file_name);
            gtk_widget_show_all(ipctx->dialog);
        }
        if (ipctx->label) {
            label = malloc(sizeof("Indexing: ") + strlen(ctx->file_name));
            if (label) {
                strcpy(label, "Indexing: ");
                strcat(label, ctx->file_name);
                gtk_label_set_text(GTK_LABEL(ipctx->label), label);
                free(label);
            }
        }
    }
    else {
        eof = mpgedit_edit_index(ctx->playctx);
    }
    if (!ipctx->quit && !eof) {
        tnow   = mpgedit_edit_index_sec(ctx->playctx);
        frames = mpgedit_edit_index_frames(ctx->playctx);
        if (tnow > ctx->tlast) {
            offset = mpgedit_edit_index_offset(ctx->playctx);
            percent = ((gdouble) offset) / ((gdouble) ctx->fsize);
            gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ipctx->progress), 
                                          percent);
            ctx->tlast = tnow + 10;
        }
    }
    else {
        if (ctx->callback) {
            ctx->callback(ctx->widget, ctx->event, ctx->data);
        }
        mpgedit_edit_index_free(ctx->playctx);        
        ctx->edit_row->indexing = FALSE;
        free(ctx);
        ipctx->ifctx            = NULL;
    }
    return 1;
}


/*
 * Callback on the indexfile dialog.  There is a cancel button on the
 * index dialog.  Note: Cancelling indexing does not cancel the need to
 * create the file index.  The next time the file is clicked on, it will
 * be indexed again, until indexing is complete.
 */
void cb_indexfile_dialog_quit(GtkWidget *widget, gint response, gpointer data)
{
    indexfile_idleproc_ctx *ctx = (indexfile_idleproc_ctx *) data;

    if (response == GTK_RESPONSE_CANCEL) {
        ctx->quit = TRUE;
    }
}


void cb_editfile_dialog_quit(GtkWidget *widget, gint response, gpointer data)
{
    idle_editfile_ctx *ctx = (idle_editfile_ctx *) data;

    if (response == GTK_RESPONSE_CANCEL) {
        ctx->quit = TRUE;
    }
}


GtkWidget *___make_indexfile_progress_dialog(indexfile_idleproc_ctx *ctx,
                                             GtkWidget *parent,
                                             char *file_name)
{
    GtkWidget *dialog;
    GtkWidget *progress;
    GtkWidget *label;
    char      *str;

    str = malloc(PATH_MAX);
    if (!str) {
        return NULL;
    }

    strcpy(str, "Indexing: ");
    strcat(str, file_name);
    dialog = gtk_dialog_new_with_buttons(
                 str, GTK_WINDOW(parent), GTK_DIALOG_MODAL,
                 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                 NULL);

    g_signal_connect(G_OBJECT(dialog), "response", 
                     G_CALLBACK(cb_indexfile_dialog_quit), ctx);
                     
    progress      = gtk_progress_bar_new();
    ctx->progress = progress;
    gtk_widget_show(progress);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
                       progress, FALSE, FALSE, 0);

    label      = gtk_label_new(str);
    ctx->label = label;
    gtk_widget_show(label);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
                       label, FALSE, FALSE, 0);

    free(str);
    return dialog;
}


GtkWidget *___make_editfile_progress_dialog(idle_editfile_ctx *ctx,
                                            GtkWidget *parent,
                                            char *file_name)
{
    GtkWidget *dialog;
    GtkWidget *progress;
    GtkWidget *label;
    char      *str;

    str = malloc(PATH_MAX);
    if (!str) {
        return NULL;
    }

    sprintf(str, "Editing: %s", file_name);
    dialog = gtk_dialog_new_with_buttons(
                 str, GTK_WINDOW(parent), GTK_DIALOG_MODAL,
                 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                 NULL);

    g_signal_connect(G_OBJECT(dialog), "response", 
                     G_CALLBACK(cb_editfile_dialog_quit), ctx);
                     
    progress = gtk_progress_bar_new();
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
                       progress, FALSE, FALSE, 0);

    label = gtk_label_new(str);
    gtk_widget_show(label);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
                       label, FALSE, FALSE, 0);

    ctx->progress = progress;
    gtk_widget_show(progress);

    free(str);
    return dialog;
}


/*
 * Determine if the index file must be created before processing the 
 * row clicked callback. mpgedit_edit_index_init() is called, and
 * when a context handle is returned, an idle proc is created to 
 * finish the indexing operation.  Once that processing is completed
 * by row_clicked_idle_indexfile(), callback(widget, event, ER) is
 * called.
 */
void ___indexfile_init(GtkWidget *widget, GdkEvent *event, edit_row_ctx *er,
                       void (*callback)(GtkWidget *, GdkEvent *, gpointer))
{
    global_vars_ctx        *gvctx;
    idle_indexfile_ctx     *ifctx;
    indexfile_idleproc_ctx *idlectx;

    if (er->indexing) {
        return;
    }
    idlectx = (indexfile_idleproc_ctx *)
                  calloc(1, sizeof (indexfile_idleproc_ctx));
    if (!idlectx) {
        return;
    }
    ifctx = (idle_indexfile_ctx *) calloc(1, sizeof(idle_indexfile_ctx));
    if (!ifctx) {
        free(idlectx);
        return;
    }
    gvctx            = (global_vars_ctx *) er->ctx;
    idlectx->gvctx   = gvctx;
    er->indexing     = TRUE;
    ifctx->widget    = widget;
    ifctx->event     = event;
    ifctx->data      = (gpointer) er;
    ifctx->file_name = er->file_name;
    ifctx->edit_row  = er;
    ifctx->callback  = callback;

    /*
     * row_clicked_idle_indexfile() extracts this entry off the
     * list for processing.
     */
    ___index_file_list_put(gvctx, ifctx);

    if (gvctx->idle_indexfile_count < MAX_IDLE_INDEX_PROC) {
        gvctx->idle_indexfile_count++;
        gtk_idle_add(row_clicked_idle_indexfile, idlectx);
    }
}


void cb_row_clicked(GtkWidget *widget,
                    int row,
                    int column,
                    GdkEventButton *event,
                    gpointer data)
{
    edit_row_ctx *er;
    global_vars_ctx *gvctx = (global_vars_ctx *) data;

    er = ___edit_row_find_index(&gvctx->editctx, row);

#ifdef _DEBUG
    printf("cb_row_clicked: row=%d column=%d er = %p\n", row, column, er);
#endif
    if (er) {
        ___indexfile_init(widget, (GdkEvent *) event, er, cb_row_clicked_1);
        gtk_widget_set_sensitive(gvctx->menu_edit_cut,       row>0);
        gtk_widget_set_sensitive(gvctx->menu_edit_copy,      row>0);
        gtk_widget_set_sensitive(gvctx->menu_edit_duplicate, row>0);
        gtk_widget_set_sensitive(gvctx->menu_edit_time,      row>1);
        gtk_widget_set_sensitive(gvctx->menu_edit,
                                 (gvctx->editctx.head != gvctx->editctx.tail));
    }
}



void cb_file_sel_ok(GtkWidget *widget, GtkFileSelection *fs)
{
    global_vars_ctx *gvctx = (global_vars_ctx *) fs;
    edit_row_ctx    *edit_row;
    GtkAdjustment   *adj;
    gdouble         incr;
    char            *item[3];

    ___file_sel_ok(widget, (gpointer) gvctx->idlectx,  0, 0);
    gvctx->eidlectx->adj_elapsed_label_noset = 1;
    ___file_sel_ok(widget, (gpointer) gvctx->eidlectx, 0, 0);
    gvctx->eidlectx->adj_elapsed_label_noset = 0;
    if (gvctx->editctx.head) {
        gvctx->editctx.dirty = TRUE;
    }
    edit_row = 
        ___row_make(gvctx->idlectx->play_file, 0, 
                    0, 0, 0, 
                    0, 0, 0, 
                    G_CALLBACK(cb_row_clicked), (gpointer) gvctx);
    ___edit_row_list_append(edit_row, &gvctx->editctx);

    adj = gtk_scrolled_window_get_vadjustment(
              GTK_SCROLLED_WINDOW(gvctx->editctx.scrolled_window));
    incr = adj->upper / ((gdouble) gvctx->editctx.tail->row_num) + 2.0;
    gtk_adjustment_set_value(adj, adj->value + incr);
    item[0] = edit_row->file_name;
    item[1] = edit_row->stime_str;
    item[2] = edit_row->etime_str;
    gtk_clist_append(GTK_CLIST(gvctx->editctx.clist), item);
    gtk_clist_select_row(GTK_CLIST(gvctx->editctx.clist), edit_row->row_num, 0);
    gvctx->editctx.cursor = edit_row;
    gtk_widget_set_sensitive(gvctx->menu_save_session, gvctx->editctx.dirty);
}


int ___test_is_valid_file_error_dialog(const char *name, global_vars_ctx *gvctx)
{
    struct stat sbuf;
    int sts;
    GtkWidget *dialog;

    sts = stat(name, &sbuf);
    if (sts == 0 && !S_ISREG(sbuf.st_mode)) {
        sts = 1;
    }
/*
 * Should do even better here, and test if file is an MPEG audio file
 */
    if (sts) {
        dialog = ___make_dialog_informational(gvctx->main_window,
                     FALSE,
                     (char *) name,
                     "        ERROR: Invalid filename specified        \n\n");
        if (dialog) {
            gtk_widget_show_all(dialog);
            gtk_dialog_run(GTK_DIALOG(dialog));
            gtk_widget_destroy(dialog);
        }
    }
    return sts;
}


void cb_file_sel_ok_1(GtkWidget *widget, GtkFileSelection *fs)
{
    global_vars_ctx *gvctx = (global_vars_ctx *) fs;
    char            *name;
    char            *cp;

    gtk_widget_hide(gvctx->fs);
    gtk_toggle_button_set_active(
        GTK_TOGGLE_BUTTON(gvctx->open_file_toggle), FALSE);
    name = (char *) gtk_file_selection_get_filename(GTK_FILE_SELECTION(gvctx->fs));

    /*
     * Map .idx->.mp3, since it is not valid to open an index file
     */
    cp = strstr(name, ".idx");
    if (cp && strlen(cp) == 4) {
        strcpy(cp, ".mp3");
    }

    /*
     * Test specified file name for validity.
     */
    if (___test_is_valid_file_error_dialog(name, gvctx)) {
        gtk_entry_set_text(GTK_ENTRY(gvctx->entry), "");
        return;
    }

    if (gvctx->idlectx->play_file) {
        free(gvctx->idlectx->play_file);
    }

    name = strdup(name);
    /*
     * Convert backslashes in path to '/'
     */
    cp = name;
    while (*cp) {
        if (*cp == '\\') {
            *cp = '/';
        }
        cp++;
    }

    gvctx->idlectx->play_file  = name;
    gvctx->eidlectx->play_file = name;
    gtk_tooltips_enable(gvctx->tooltips_entry);
    gtk_tooltips_set_tip(gvctx->tooltips_entry, GTK_WIDGET(gvctx->entry),
                         name, NULL);

    gtk_entry_set_text(GTK_ENTRY(gvctx->entry), name);
    gtk_editable_set_position(
        GTK_EDITABLE(gvctx->entry), strlen(name));

    cb_file_sel_ok(widget, fs);
}



void cb_entry(GtkWidget *widget, void *data)
{
    global_vars_ctx *gvctx = (global_vars_ctx *) data;
    const char      *name;
    char            *cp;

    name = gtk_entry_get_text(GTK_ENTRY(widget));
#ifdef _DEBUG
    printf("cb_entry: called '%s'\n", name);
#endif
    /*
     * Map .idx->.mp3, since it is not valid to open an index file
     */
    cp = strstr(name, ".idx");
    if (cp && strlen(cp) == 4) {
        strcpy(cp, ".mp3");
    }
    if (___test_is_valid_file_error_dialog(name, gvctx)) {
        gtk_entry_set_text(GTK_ENTRY(widget), "");
        return;
    }
    if (gvctx->idlectx->play_file) {
        free(gvctx->idlectx->play_file);
    }
    name = strdup((char *) name);
    gvctx->idlectx->play_file  = (char *) name;
    gvctx->eidlectx->play_file = (char *) name;

    cb_file_sel_ok(widget, data);
}


void cb_file_sel_cancel(GtkWidget *widget, GtkFileSelection *fs)
{
    global_vars_ctx *gvctx = (global_vars_ctx *) fs;

    gtk_widget_hide(gvctx->fs);
    gtk_toggle_button_set_active(
        GTK_TOGGLE_BUTTON(gvctx->open_file_toggle), FALSE);
}


void cb_open_file_from_menu(gpointer data)
{
    global_vars_ctx *gvctx = (global_vars_ctx *) data;

    gtk_toggle_button_set_active(
        GTK_TOGGLE_BUTTON(gvctx->open_file_toggle), TRUE);

    gtk_widget_show(gvctx->fs);
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(gvctx->fs), "");
}


void cb_open_file(GtkWidget *widget, gpointer data)
{
    global_vars_ctx *gvctx = (global_vars_ctx *) data;
    int active  = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gvctx->open_file_toggle));

    if (active) {
        gtk_widget_show(gvctx->fs);
        gtk_file_selection_set_filename(GTK_FILE_SELECTION(gvctx->fs), "");
    }
    else {
        gtk_widget_hide(gvctx->fs);
    }
}


gint cb_delete(GtkWidget *widget,
               GdkEvent  *event,
               gpointer   data )
{
    idle_ctx *idlectx = (idle_ctx *) data;
    int playing = gtk_toggle_button_get_active(
                      GTK_TOGGLE_BUTTON(idlectx->play_toggle));

    if (playing) {
        mpgedit_play_close(idlectx->playctx);
        idlectx->playctx = NULL; 
    }
    gtk_main_quit();
    return FALSE;
}


void ___list_iterator_func(void *ctx, edit_row_ctx *row)
{
    cb_edit_ctx *edctx = (cb_edit_ctx *) ctx;
    char timespec[128];

    sprintf(timespec, "%d:%d.%d-%d:%d.%d",
            row->s_min, row->s_sec, row->s_msec,
            row->e_min, row->e_sec, row->e_msec);
            
    if (row->file_name && row->file_name[0]) {
        mpgedit_editspec_append(edctx->edarray, row->file_name, timespec);
    }
#if 1 /* is this needed?? */
    edctx->index++;
#endif
}


long ___get_file_total_secs(char *file)
{
    void *ctx;
    long rsecs = -1;

    ctx = mpgedit_play_init(file);
    if (ctx) {
        rsecs = mpgedit_play_total_sec(ctx);
        mpgedit_play_close(ctx);
    }
    return rsecs;
}


long ___compute_edit_total_secs(editspec_t *edarray, int edlen)
{
    long      ssec;
    long      susec;
    long      esec;
    long      eusec;
    char      *filename;
    long rtsecs = 0;
    long diffsecs;
    long esecs;
    int  i;

    for (i=0; i<edlen; i++) {
        mpeg_time_gettime(mpgedit_editspec_get_stime(edarray, i),
                          &ssec, &susec);
        mpeg_time_gettime(mpgedit_editspec_get_etime(edarray, i),
                          &esec, &eusec);
        filename = mpgedit_editspec_get_file(edarray, i);
        if (esec == MP3_TIME_INFINITE && eusec == 0) {
            esecs = ___get_file_total_secs(filename);
        }
        else {
            esecs = esec;
        }
        diffsecs = esecs - ssec;
        if (diffsecs > 0) {
            rtsecs += diffsecs;
        }
    }
    return rtsecs;
}


#if _DEBUG
static void debug_print_editspec(editspec_t *edarray)
{
    editspec_t *edentry;
    int        indx;
    int        len;
    int        arraylen;
    int        i;
    char       *filename;
    long       ssec;
    long       susec;
    long       esec;
    long       eusec;

    arraylen = mpgedit_editspec_get_length(edarray);
    indx = 0;
    while (indx<arraylen) {
        edentry = mpgedit_editspec_get_edit(edarray, indx,
                                            arraylen, 0, &len);
        {
            for (i=0; i<len; i++) {
                filename = mpgedit_editspec_get_file(edentry, i);
                mpeg_time_gettime(mpgedit_editspec_get_stime(edentry, i),
                                  &ssec, &susec);
                mpeg_time_gettime(mpgedit_editspec_get_etime(edentry, i),
                                  &esec, &eusec);
                printf("%s (%ld:%ld-%ld:%ld)\n",
                       filename, ssec, susec, esec, eusec);
            }
            printf("\n");
        }
        indx += len;
    }
}
#endif


gint cb_edit_idle_editfile(gpointer data)
{
    idle_editfile_ctx *ctx = (idle_editfile_ctx *) data;
    gint              rval = TRUE;
    int               sts = 0;
    int               built_len;
    gdouble           percent;
    GtkWidget         *dialog;
    int               append    = FALSE;
    int               *indexptr = NULL;
    char              *filename;
    long              ssec;
    long              susec;
    long              esec;
    long              eusec;

    eusec = esec = susec = ssec = 0;
    filename = NULL;
    if (ctx->gvctx->ofdialogctx->action == OUTPUT_FILE_APPEND_JOIN) {
        append = TRUE;
    }
    else {
        indexptr = &ctx->out_fileindx;
    }


    
    /*
     * Iterate over the split operation.  Must first initialize the
     * context values, then perform the remaining edit operations
     * iteratively.  Since there are two loops being maintained,
     * a state machine is used to keep everything straight.
     */
    switch (ctx->state) {
      case EDIT_IDLE_STATE_INIT:
        ctx->edit_indx    = 0;
        ctx->edit_flags = append ? MPGEDIT_FLAGS_APPEND : 0;
        ctx->out_fileindx = 1;
        ctx->state = EDIT_IDLE_STATE_LOOP;
#if _DEBUG
        debug_print_editspec(ctx->edarray);
#endif
        break;

      case EDIT_IDLE_STATE_LOOP:
        if (ctx->edit_indx >= ctx->edarray_len) {
            ctx->state = EDIT_IDLE_STATE_DONE;
            break;
        }
        ctx->current_sec   = 0;
        if (ctx->built_edarray) {
            free(ctx->built_edarray);
        }
#if 1
        ctx->built_edarray =  
            mpgedit_editspec_get_edit(ctx->edarray, ctx->edit_indx,
                                      ctx->edarray_len, append, &built_len);
#else
        ctx->built_edarray = 
            mpgedit_editspec_get_edit(ctx->edarray, ctx->edit_indx,
                ctx->edarray_len - ctx->edit_indx,
                append, &built_len);
#endif

        if (!ctx->built_edarray) {
            /* Malloc error */
            ctx->edit_indx = ctx->edarray_len;
            ctx->state     = EDIT_IDLE_STATE_DONE;
            break;
        }

        if (ctx->out_filename) {
            free(ctx->out_filename);
        }
        ___make_default_output_filename(ctx->gvctx);
        ctx->out_filename = mpgedit_edit_make_output_filename(
                                NULL,
                                ctx->gvctx->output_file,
                                "mp3",
                                indexptr);
#if defined(_DEBUG)
        {
            int i;
            printf("starting edits %d output=%s\n",
                   ctx->edit_indx, ctx->out_filename);
            for (i=0; i<built_len; i++) {
                filename = mpgedit_editspec_get_file(ctx->built_edarray, i);
                mpeg_time_gettime(mpgedit_editspec_get_stime(
                                      ctx->built_edarray, i),
                                  &ssec, &susec);
                mpeg_time_gettime(mpgedit_editspec_get_etime(
                                      ctx->built_edarray, i),
                                  &esec, &eusec);
                printf("%s (%ld:%ld-%ld:%ld)\n", 
                       filename, ssec, susec, esec, eusec);
            }
        }
#endif
        ctx->total_sec = ___compute_edit_total_secs(ctx->built_edarray,
                                                    built_len);
        ctx->edit_indx += built_len;
        ctx->edfiles_ctx = mpgedit_edit_files_init(
                               ctx->built_edarray,
                               ctx->out_filename,
                               ctx->edit_flags,
                               &sts);
        if (ctx->edfiles_ctx && sts == 0) {
            ctx->edfiles_ctx_editing = 
                mpgedit_edit_files(ctx->edfiles_ctx, &sts);
        }
        if (sts) {
            ctx->state = EDIT_IDLE_STATE_ERROR;
        }
        else if (!ctx->edfiles_ctx_editing) {
            ctx->state = EDIT_IDLE_STATE_LOOP;
            mpgedit_edit_files_free(ctx->edfiles_ctx);
        }
        else {
            ctx->state = EDIT_IDLE_STATE_EDIT;
            if (ctx->dialog) {
                gtk_widget_destroy(ctx->dialog);
            }
            ctx->dialog = ___make_editfile_progress_dialog(
                              ctx,
                              ctx->gvctx->main_window,
                              ctx->out_filename);
            gtk_widget_show_all(ctx->dialog);
        }
        break;

      case EDIT_IDLE_STATE_EDIT:
        if (ctx->quit) {
            ctx->state = EDIT_IDLE_STATE_DONE;
            break;
        }
        ctx->current_sec++;
        ctx->edfiles_ctx_editing = 
            mpgedit_edit_files(ctx->edfiles_ctx, &sts);
        percent = ((gdouble)ctx->current_sec) / ((double)ctx->total_sec);
        if (percent > 1.0) {
            percent = 1.0;
        }
        gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(ctx->progress), 
                                      percent);
        
        if (!ctx->edfiles_ctx_editing) {
            ctx->state = EDIT_IDLE_STATE_LOOP;
            mpgedit_edit_files_free(ctx->edfiles_ctx);
        }
        if (sts) {
            ctx->state = EDIT_IDLE_STATE_ERROR;
        }
        break;

      case EDIT_IDLE_STATE_DONE:
        rval = FALSE;
        if (ctx->dialog) {
            gtk_widget_destroy(ctx->dialog);
            ctx->dialog = NULL;
        }

        /* adam/TBD: free idle context memory here */
        break;

      case EDIT_IDLE_STATE_ERROR:
        if (ctx->dialog) {
            gtk_widget_destroy(ctx->dialog);
            ctx->dialog = NULL;
        }
        ctx->quit = TRUE;
        dialog = ___make_dialog_informational(ctx->gvctx->main_window,
                     FALSE,
                     ctx->out_filename,
                     "        ERROR: Edit output "
                     "file already exists       \n\n");
        if (dialog) {
            gtk_widget_show_all(dialog);
            gtk_dialog_run(GTK_DIALOG(dialog));
            gtk_widget_destroy(dialog);
            ctx->dialog = NULL;
        }
        rval = FALSE;
        ctx->state = EDIT_IDLE_STATE_INIT;
        break;
    }

    return rval;
}


void ___load_editor_by_editspec(global_vars_ctx *gvctx,
                                editspec_t *edarray,
                                int edlen)
{
    int             i;
    int             j;
    edit_row_ctx    *edit_row;
    edit_row_ctx    *cursor = NULL;
    int             found;
    char            *item[3];
    char            *filename;
    long            ssec;
    long            susec;
    long            esec;
    long            eusec;

    if (gvctx->idlectx->play_file) {
        free(gvctx->idlectx->play_file);
    }
    gvctx->idlectx->play_file  = strdup(mpgedit_editspec_get_file(edarray, 0));
    gvctx->eidlectx->play_file = gvctx->idlectx->play_file;
    gtk_widget_hide(gvctx->editctx.clist);
    for (i=0; i<edlen; i++) {
        filename = mpgedit_editspec_get_file(edarray, i);
        mpeg_time_gettime(mpgedit_editspec_get_stime(edarray, i),
                          &ssec, &susec);
        mpeg_time_gettime(mpgedit_editspec_get_etime(edarray, i),
                          &esec, &eusec);
        if (esec == MP3_TIME_INFINITE) {
            esec = 0;
        }
        edit_row = 
            ___row_make(filename, i,
                        ssec/60,
                        ssec%60,
                        susec/1000,
                        esec/60,
                        esec%60,
                        eusec/1000,
                        G_CALLBACK(cb_row_clicked), (gpointer) gvctx);
        ___edit_row_list_append(edit_row, &gvctx->editctx);
        item[0] = edit_row->file_name;
        item[1] = edit_row->stime_str;
        item[2] = edit_row->etime_str;
        gtk_clist_append(GTK_CLIST(gvctx->editctx.clist), item);
        if (!cursor) {
            cursor = edit_row;
        }
        /*
         * Don't index any file name already present in edit array
         */
        found = 0;
        for (j=0; j<i; j++) {
            if (!strcmp(mpgedit_editspec_get_file(edarray, j), 
                        mpgedit_editspec_get_file(edarray, i)))
            {
                found = 1;
            }
        }
        if (!found) {
            ___indexfile_init(NULL, NULL, edit_row, cb_row_clicked_1);
        }
    }
    gtk_widget_show(gvctx->editctx.clist);
    gvctx->editctx.cursor = cursor;
}


void ___load_editor_by_files(global_vars_ctx *gvctx, char **files, int len)
{
    int        i;
    editspec_t *edarray;

    edarray = mpgedit_editspec_init();
    if (!edarray) {
        return;
    }

    for (i=0; i<len; i++) {
        mpgedit_editspec_append(edarray, files[i], "0:0.0-0:0.0");
    }
    ___load_editor_by_editspec(gvctx, edarray, len);
}


void cb_load(GtkWidget *widget, gint response, gpointer data)
{
    global_vars_ctx *gvctx = (global_vars_ctx *) data;
    editspec_t      *edarray;
    int             edlen;
    GtkAdjustment   *adj;
    GtkWidget       *dialog;

    if (response == GTK_RESPONSE_OK) {
        edlen = ___read_abandoned_edits(gvctx->abandon_file, &edarray);
        if (edlen > 0) {
            if (gvctx->editctx.head && gvctx->editctx.head->next) {
                /*
                 * First line of editor is always the pad blank line, so
                 * dirty the editor when there are lines after this pad.
                 */
                gvctx->editctx.dirty = TRUE;
                gtk_widget_set_sensitive(gvctx->menu_save_session, TRUE);
            }
            ___load_editor_by_editspec(gvctx, edarray, edlen);
            
            gtk_clist_select_row(GTK_CLIST(gvctx->editctx.clist),
                                 gvctx->editctx.cursor->row_num, 0);
            if (gvctx->editctx.cursor->row_num > 0) {
                adj = gtk_scrolled_window_get_vadjustment(
                          GTK_SCROLLED_WINDOW(gvctx->editctx.scrolled_window));
                gtk_adjustment_set_value(adj, adj->upper);
            }
        }
        else {
           dialog = ___make_dialog_informational(
                        gvctx->main_window,
                        FALSE, gvctx->abandon_file, 
                        "        ERROR: Failed opening "
                        "saved session file        \n\n");
           gtk_widget_show_all(dialog);
           gtk_dialog_run(GTK_DIALOG(dialog));
           gtk_widget_destroy(dialog);
        }
    }
}



void cb_edit(GtkWidget *widget, gpointer data)
{
    global_vars_ctx   *gvctx = (global_vars_ctx *) data;
    int               playing = 0;
    editspec_t        *edarray;
    cb_edit_ctx       edctx;
    int               edarray_len;
    int               edit_flags = 0;
    char              *saved_edits_file;
    idle_editfile_ctx *edidle_ctx;
    GtkWidget         *dialog;

    if (!gvctx->editctx.head) {
        return;
    }
    edarray_len = gvctx->editctx.tail->row_num;

    edarray = mpgedit_editspec_init();
    if (!edarray) {
        return;
    }

    memset(&edctx, 0, sizeof(edctx));
    edctx.len = edarray_len;
    edctx.edarray = edarray;

    if (gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(gvctx->idlectx->play_toggle)))
    {
        playing |= 1;
    }
#if 0
    /*
     * adam/TBD:  This is bogus.  Not able to deactivate the
     * end play control in the same manner as the play control.
     * This needs further investigation.
     */
    else if (gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(gvctx->eidlectx->play_toggle)))
    {
        playing |= 1;
    }
#endif
 
    if (playing) {
        mpgedit_play_close(gvctx->idlectx->playctx);
        gvctx->idlectx->playctx = NULL; 
    }

    ___edit_row_list_iterate(&gvctx->editctx, ___list_iterator_func, &edctx);

    /*
     * Overwrite of abandoned edits file has already been
     * accepted by query_abandon_ctx; see cb_edit_1().
     */
    saved_edits_file = gvctx->abandon_file;
    if (!___write_abandoned_edits(saved_edits_file, 1,
                                  edarray, edarray_len))
    {
        dialog = ___make_dialog_informational(gvctx->main_window,
                     FALSE,
                     saved_edits_file,
                     "        ERROR: Abandoned session "
                     "save file already exists        \n\n");
        gtk_widget_show_all(dialog);
        gtk_dialog_run(GTK_DIALOG(dialog));
        gtk_widget_destroy(dialog);
        gvctx->edit_action_abandon_save = FALSE;
        return;
    }
    gvctx->editctx.dirty = FALSE;
    gtk_widget_set_sensitive(gvctx->menu_save_session, FALSE);

    /* 
     * Action was only to save abandon edits when set
     */
    if (gvctx->edit_action_abandon_save) {
        gvctx->edit_action_abandon_save = FALSE;
        return;
    }

    /*
     * All of the file split editing is performed in the 
     * idle proc function cb_edit_idle_editfile().
     */
    edidle_ctx = calloc(1, sizeof(*edidle_ctx));
    if (edidle_ctx) {
        edidle_ctx->gvctx       = gvctx;
        edidle_ctx->edfiles_ctx = NULL;
        edidle_ctx->edfiles_ctx_editing = 0;
        edidle_ctx->edarray     = edarray;
        edidle_ctx->edarray_len = edarray_len;
        edidle_ctx->edit_flags  = edit_flags;
        gtk_idle_add(cb_edit_idle_editfile, edidle_ctx);
    }
}



/*
 * This layer is needed because we are called by ___indexfile_init,
 * which passes as 3 arguments.  cb_edit() only takes widget, and user
 * data, so we must "eat" the response argument
 */
void cb_edit_3(GtkWidget *widget, GdkEvent *dummy, gpointer data)
{
    edit_row_ctx *er = (edit_row_ctx *) data;

    cb_edit(widget, er->ctx);
}


void cb_edit_1_rmd(gpointer data)
{
    void cb_edit_1a(GtkWidget *widget, gpointer data);
    global_vars_ctx *gvctx = (global_vars_ctx *) data;

    cb_edit_1a(gvctx->row_modified_ctx.widget, data);
}


void cb_edit_1a(GtkWidget *widget, gpointer data)
{
    global_vars_ctx   *gvctx = (global_vars_ctx *) data;
    int               len;
    editspec_t        *edarray;
    edit_row_ctx      *er;

    /*
     * Test existence of abandoned session file, and display warning
     * dialog when already exists.
     */
    len = ___read_abandoned_edits(gvctx->abandon_file, &edarray);
    if (len > 0 && gvctx->editctx.dirty) {
        free(edarray);
        gtk_widget_show_all(gvctx->query_abandon_ctx->dialog);
    }
    else {
        er = (edit_row_ctx *) calloc(1, sizeof(edit_row_ctx));
        if (er) {
            er->ctx = gvctx;
            ___indexfile_init(widget, NULL, er, cb_edit_3);
        }
    }
}


void cb_edit_1(GtkWidget *widget, gpointer data)
{
    global_vars_ctx   *gvctx = (global_vars_ctx *) data;

    /*
     * Do nothing when action is save session, but buffer is not dirty.
     */
    if (gvctx->edit_action_abandon_save && !gvctx->editctx.dirty) {
        gvctx->edit_action_abandon_save = FALSE;
        return;
    }

    /*
     * Prompt to save changes to current row
     */
    if (gvctx->idlectx->dirty || gvctx->eidlectx->dirty) {
        gvctx->row_modified_ctx.proc   = cb_edit_1_rmd;
        gvctx->row_modified_ctx.data   = gvctx;
        gvctx->row_modified_ctx.widget = widget;
        ___row_modified_dialog(gvctx);
        return;
    }

    cb_edit_1a(widget, gvctx);
}


void cb_edit_2(GtkWidget *widget, gint response,  gpointer data)
{
    global_vars_ctx   *gvctx = (global_vars_ctx *) data;

    if (response == GTK_RESPONSE_OK) {
        ___put_default_abandoned_filename(gvctx->abandon_file);
        cb_edit(widget, data);
    }
}


gint cb_quit(GtkWidget *widget, gpointer data)
{
    idle_ctx  *idlectx = (idle_ctx *) data;
    gint      response;
    int       playing = gtk_toggle_button_get_active(
                            GTK_TOGGLE_BUTTON(idlectx->play_toggle));

    if (playing) {
        mpgedit_play_close(idlectx->playctx);
        idlectx->playctx = NULL; 
    }

    if (idlectx->gvctx->editctx.dirty) {
        response = ___editor_dirty_dialog_run(idlectx->gvctx);
        if (response != GTK_RESPONSE_OK) {
            return FALSE;
        }
    }

    gtk_main_quit();
    return TRUE;
}


void cb_pause_button(GtkWidget *widget, gpointer data)
{
    idle_ctx *idlectx = (idle_ctx *) data;
    int playing;
    int paused;


    if (!___button_used_match(widget, idlectx->gvctx->pause_toggle, idlectx)) {
        return;
    }

    paused = gtk_toggle_button_get_active(
                 GTK_TOGGLE_BUTTON(idlectx->pause_toggle));
    playing = gtk_toggle_button_get_active(
                  GTK_TOGGLE_BUTTON(idlectx->play_toggle));
    if (playing) {
        if (paused) {
            idlectx->pause_sec  = mpgedit_play_current_sec(idlectx->playctx);
            idlectx->pause_usec = mpgedit_play_current_msec(idlectx->playctx);
            mpgedit_play_close(idlectx->playctx);
            idlectx->playctx = NULL;
            gtk_idle_remove(idlectx->idle_tag);
        }
        else {
            idlectx->playctx = mpgedit_play_init(idlectx->play_file);
            if (!idlectx->playctx) {
                ___make_cant_open_sound(idlectx->gvctx);
               return;
            }
            mpgedit_play_set_status_callback(idlectx->playctx,
                                             cb_label_set_text, 
                                             idlectx);
            mpgedit_play_seek_time(idlectx->playctx,
                                   idlectx->pause_sec, idlectx->pause_usec);
            idlectx->idle_tag = gtk_idle_add(idle_play, idlectx);
        }
    }
    else {
        if (paused) {
            gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(idlectx->pause_toggle), FALSE);
            ___button_used_clear(idlectx);
        }
    }
}


void cb_stop_button(GtkWidget *widget, gpointer data)
{
    idle_ctx *idlectx = (idle_ctx *) data;

    /* Must push button out because it is a toggle button */
    gtk_toggle_button_set_active(
        GTK_TOGGLE_BUTTON(idlectx->stop_toggle), FALSE);
    if (!___button_used_match(widget, idlectx->gvctx->stop_toggle, idlectx)) {
        return;
    }

    if (gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(idlectx->playto_toggle))) 
    {
        gtk_toggle_button_set_active(
            GTK_TOGGLE_BUTTON(idlectx->playto_toggle), FALSE);
    }
    else if (gtk_toggle_button_get_active(
            GTK_TOGGLE_BUTTON(idlectx->play_toggle)))
    {
        gtk_toggle_button_set_active(
            GTK_TOGGLE_BUTTON(idlectx->play_toggle), FALSE);
    }
    ___button_used_clear(idlectx);
}


void cb_play_button(GtkWidget *widget, gpointer data)
{
    idle_ctx *idlectx = (idle_ctx *) data;
    int playing;
    long sec;
    long usec;
    long esec;
    long eusec;


    if (!___button_used_match(widget, idlectx->gvctx->play_toggle, idlectx)) {
        return;
    }

    /*
     * Quit if playto is running
     */
    playing = gtk_toggle_button_get_active(
                  GTK_TOGGLE_BUTTON(idlectx->playto_toggle));
    if (playing) {
        gtk_toggle_button_set_active(
            GTK_TOGGLE_BUTTON(idlectx->play_toggle), FALSE);
        return;
    }

    playing = gtk_toggle_button_get_active(
                  GTK_TOGGLE_BUTTON(idlectx->play_toggle));
    if (playing) {
        /*
         * Transitioned from stopped -> playing
         */
        if (!idlectx->play_file ||
            ___test_is_valid_file_error_dialog(idlectx->play_file,
                                               idlectx->gvctx))
        {
            gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(idlectx->play_toggle), FALSE);
            ___button_used_clear(idlectx);
            return;
        }
        idlectx->playctx = mpgedit_play_init(idlectx->play_file);
        if (!idlectx->playctx) {
            ___make_cant_open_sound(idlectx->gvctx);
            /*
             * Do something more exciting when this fails
             */
            gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(idlectx->play_toggle), FALSE);
            ___button_used_clear(idlectx);
            return;
        }
        mpgedit_play_seek_time(idlectx->playctx, 
                               idlectx->stime_sec, idlectx->stime_usec);
#ifdef _DEBUG
        printf("cb_play_button: "
               "stime=%d:%d | elapsed=%d:%d size=%ld sec=%ld usec=%ld\n",
               idlectx->stime_sec, idlectx->stime_usec,
               idlectx->elapsed_sec, idlectx->elapsed_usec,
               mpgedit_play_total_size(idlectx->playctx),
               mpgedit_play_total_sec(idlectx->playctx),
               mpgedit_play_total_msec(idlectx->playctx));
#endif
        idlectx->idle_tag = gtk_idle_add(idle_play, idlectx);
        mpgedit_play_set_status_callback(idlectx->playctx,
                                         cb_label_set_text, 
                                         idlectx);
    }
    else {
        /*
         * Transitioned from playing -> stopped
         */
        if (idlectx->play_toggle_ignore) {
            return;
        }
        idlectx->elapsed_sec  = idlectx->stime_sec;
        idlectx->elapsed_usec = idlectx->stime_usec;
        label_set_text(idlectx->label_elapsed, "%d:%02d.%03d (%ds)", 
                       idlectx->stime_sec/60, 
                       idlectx->stime_sec%60, 
                       idlectx->stime_usec,
                       idlectx->stime_sec);

        mpgedit_play_close(idlectx->playctx);
        idlectx->playctx = NULL; 
        if (idlectx->idle_tag) {
            gtk_idle_remove(idlectx->idle_tag);
        }
        gtk_toggle_button_set_active(
            GTK_TOGGLE_BUTTON(idlectx->pause_toggle), FALSE);

        idlectx->ignore_adjustment_set = 1;
        ___spin_button_set_value(idlectx,
                                 idlectx->stime_sec,
                                 idlectx->stime_usec);
        idlectx->ignore_spin_button_set = 1;
        ___spin_scrollbar_set_value(idlectx);
        idlectx->ignore_spin_button_set = 0;
        ___button_used_clear(idlectx);
        if (idlectx->play_toggle == widget) {
            /*
             * Set the end time to the start time when the end time
             * is less than the start time.  This constraint is required
             * because it is illegal to edit with the end time before
             * the start time.
             */
            sec   = idlectx->stime_sec;
            usec  = idlectx->stime_usec;
            esec  = idlectx->gvctx->eidlectx->stime_sec;
            eusec = idlectx->gvctx->eidlectx->stime_usec;
            if ((esec > 0 || eusec > 0) &&
                (esec < sec || (esec == sec && eusec < usec)))
            {
                idlectx->gvctx->eidlectx->stime_sec  = sec;
                idlectx->gvctx->eidlectx->stime_usec = usec;
                ___spin_button_set_value(idlectx->gvctx->eidlectx, sec, usec);
                idlectx->ignore_spin_button_set = 1;
                ___spin_scrollbar_set_value(idlectx->gvctx->eidlectx);
                idlectx->ignore_spin_button_set = 0;
            }
        }
        idlectx->ignore_adjustment_set = 0;
    }
    return;
}


GtkWidget *xpm_pixmap_box(global_vars_ctx *gvctx, gchar *xpm_filename)
{
     GtkWidget *box;
     GtkWidget *image;
     char      *xpm_file;

     xpm_file = gtk_rc_find_pixmap_in_path(gvctx->settings, NULL, xpm_filename);
     if (!xpm_file) {
         xpm_file = strdup(xpm_filename);
     }
#ifdef _DEBUG
     printf("xpm_pixmap_box: %s=<%s>\n", xpm_filename, xpm_file);
#endif

     /* Create box for image and label */
     box = gtk_hbox_new(FALSE, 0);
     gtk_container_set_border_width(GTK_CONTAINER (box), 2);

     /* Now on to the image stuff */
     image = gtk_image_new_from_file(xpm_file);
#ifndef WIN32
     /*
      * adam/TBD: Hmm, the absence of g_free on Win32 makes me wonder.
      * Calling free() crashes on Win32.  Maybe the memory being allocated in
      * one DLL and being freed by another is the problem.  Don't know.  This
      * will just have to be a memory leak for now.
      */
     g_free(xpm_file);
#endif

     /* Pack the image into the box */
     gtk_box_pack_start(GTK_BOX(box), image, FALSE, FALSE, 3);

     gtk_widget_show(image);

     return box;
}


void __construct_playback_control(idle_ctx *idlectx,
                                  global_vars_ctx *gvctx,
                                  GtkWidget *box,
                                  char *labelstr,
                                  long sec,
                                  long usec)
{
    GtkWidget *label;
    GtkWidget *box_buttons;
    GtkWidget *box_label;
    GtkWidget *box_spin;
    GtkWidget *button;
    GtkWidget *imgbox;
    GtkWidget *scrollbar;

    GtkObject *spinadj;
    GtkWidget *spin;

    box_buttons = gtk_hbox_new(FALSE, 5);
    gtk_widget_show(box_buttons);

    /*
     * start label
     */
    box_label = gtk_hbox_new(FALSE, 5);
    gtk_widget_show(box_label);


    label = gtk_label_new(labelstr);
    gtk_widget_show(label);
    gtk_box_pack_start(GTK_BOX(box_label), label, FALSE, FALSE, 3);

    label = gtk_label_new("start_value");
    idlectx->label_start = label;
    gtk_widget_show(label);
    gtk_box_pack_start(GTK_BOX(box_label), label, FALSE, FALSE, 0);
    label_set_text(idlectx->label_start, "%d:%02d (%ds)", 0, 0, 0);


    gtk_box_pack_start(GTK_BOX(box), box_label, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(box), box_buttons, FALSE, FALSE, 0);


    /*
     * Scroll bar
     */
    idlectx->adj = gtk_adjustment_new(0.0, 0.0, 
                                     ((gdouble) sec) + 
                                     ((gdouble) usec/1000.0),
                                      0.001, 0.005, 0.001);
    scrollbar = gtk_hscrollbar_new(GTK_ADJUSTMENT(idlectx->adj));
    gtk_widget_show(scrollbar);
    idlectx->scrollbar = scrollbar;
    gtk_box_pack_start(GTK_BOX(box), scrollbar, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(idlectx->adj), "value_changed",
                     G_CALLBACK(cb_scrollbar_adj), idlectx);
    g_signal_connect(G_OBJECT(idlectx->adj), "value_changed",
                     G_CALLBACK(cb_spin_modified), idlectx);

    
    /*
     * Time spin boxes
     */
    box_spin = gtk_hbox_new(FALSE, 0);
    gtk_widget_show(box_spin);

    spinadj = gtk_adjustment_new(0.0, 0.0, 9999.0, 1.0, 2.0, 0.0);
    spin = gtk_spin_button_new(GTK_ADJUSTMENT(spinadj), 0.005, 0);
    gtk_widget_show(spin);
    gtk_box_pack_start(GTK_BOX(box_spin), spin, FALSE, FALSE, 0);
    idlectx->sbctx.minute = spin;
    g_signal_connect(G_OBJECT(spin), "value_changed",
                     G_CALLBACK(cb_spin_minute), idlectx);
    g_signal_connect(G_OBJECT(spin), "value_changed",
                     G_CALLBACK(cb_spin_modified), idlectx);


    spinadj = gtk_adjustment_new(0.0, -1.0, 61.0, 1.0, 2.0, 0.0);
    spin = gtk_spin_button_new(GTK_ADJUSTMENT(spinadj), 0.05, 0);
    gtk_widget_show(spin);
    gtk_box_pack_start(GTK_BOX(box_spin), spin, FALSE, FALSE, 0);
    idlectx->sbctx.second = spin;
    g_signal_connect(G_OBJECT(spin), "value_changed",
                     G_CALLBACK(cb_spin_second), idlectx);
    g_signal_connect(G_OBJECT(spin), "value_changed",
                     G_CALLBACK(cb_spin_modified), idlectx);
    
    spinadj = gtk_adjustment_new(0.0, -1.0, 1001.0, 5.0, 100.0, 0.0);
    spin = gtk_spin_button_new(GTK_ADJUSTMENT(spinadj), 0.1, 0);
    gtk_widget_show(spin);
    gtk_box_pack_start(GTK_BOX(box_spin), spin, FALSE, FALSE, 0);
    idlectx->sbctx.millisecond = spin;
    g_signal_connect(G_OBJECT(spin), "value_changed",
                     G_CALLBACK(cb_spin_millisecond), idlectx);
    g_signal_connect(G_OBJECT(spin), "value_changed",
                     G_CALLBACK(cb_spin_modified), idlectx);
    
    /*
     * Before Play button
     */
    button = gtk_toggle_button_new();
    idlectx->playto_toggle = button;
    imgbox = xpm_pixmap_box(gvctx, IMAGE_PLAYTO);
    gtk_container_add(GTK_CONTAINER(button), imgbox);
    gtk_widget_show(button);
    gtk_widget_show(imgbox);
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(cb_playto_button), idlectx);
    gtk_box_pack_start(GTK_BOX(box_buttons), button, FALSE, FALSE, 0);
    gtk_tooltips_set_tip(gvctx->tooltips, GTK_WIDGET(button),
                         TIP_PLAYTO, NULL);

    /*
     * Pause button.  Displaying this item is configurable from the
     * options menu.
     */
    button = gtk_toggle_button_new();
    idlectx->pause_toggle = button;
    imgbox = xpm_pixmap_box(gvctx, IMAGE_PAUSE);
    gtk_container_add(GTK_CONTAINER(button), imgbox);
    gtk_widget_show(imgbox);
    g_signal_connect(G_OBJECT(button), "clicked", 
                     G_CALLBACK(cb_pause_button), idlectx);
    gtk_tooltips_set_tip(gvctx->tooltips, GTK_WIDGET(button),
                         TIP_PAUSE, NULL);
    gtk_box_pack_start(GTK_BOX(box_buttons), button, FALSE, FALSE, 0);

    /*
     * Stop button.  Displaying this item is configurable from the
     * options menu.
     */
    button = gtk_toggle_button_new();
    idlectx->stop_toggle = button;
    imgbox = xpm_pixmap_box(gvctx, IMAGE_STOP);
    gtk_container_add(GTK_CONTAINER(button), imgbox);
    gtk_widget_show(imgbox);
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(cb_stop_button), idlectx);
    gtk_tooltips_set_tip(gvctx->tooltips, GTK_WIDGET(button),
                         TIP_STOP, NULL);
    gtk_box_pack_start(GTK_BOX(box_buttons), button, FALSE, FALSE, 0);

    /*
     * Play button
     */
    button = gtk_toggle_button_new();
    idlectx->play_toggle = button;
    imgbox = xpm_pixmap_box(gvctx, IMAGE_PLAY);
    gtk_container_add(GTK_CONTAINER(button), imgbox);
    gtk_widget_show(button);
    gtk_widget_show(imgbox);
    g_signal_connect(G_OBJECT(button), "clicked", 
                     G_CALLBACK(cb_play_button), idlectx);
    gtk_box_pack_start(GTK_BOX(box_buttons), button, FALSE, FALSE, 0);
    gtk_tooltips_set_tip(gvctx->tooltips, GTK_WIDGET(button),
                         TIP_PLAY, NULL);

    /*
     * Record button
     */
    button = gtk_button_new();
    idlectx->record = button;
    imgbox = xpm_pixmap_box(gvctx, IMAGE_RECORD);
    gtk_container_add(GTK_CONTAINER(button), imgbox);
    gtk_widget_show(button);
    gtk_widget_show(imgbox);
    gtk_box_pack_start(GTK_BOX(box_buttons), button, FALSE, FALSE, 0);

    g_signal_connect(G_OBJECT(button), "clicked", 
                     G_CALLBACK(cb_record_button), gvctx);
    gtk_tooltips_set_tip(gvctx->tooltips, GTK_WIDGET(button),
                         TIP_RECORD, NULL);

    gtk_box_pack_start(GTK_BOX(box_buttons), box_spin, FALSE, FALSE, 0);
}



void cb_window_button_press(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    global_vars_ctx *gvctx         = (global_vars_ctx *) data;
    GdkEventButton *event_button = (GdkEventButton *) event;

#ifdef _DEBUG
    printf("cb_window_button_press: called\n");
#endif
    if (event->type == GDK_BUTTON_PRESS) {
#ifdef _DEBUG
        printf("cb_window_button_press: event_button: %d\n",
               event_button->button);
#endif
        if (event_button->button == 3) {
            gtk_menu_popup (GTK_MENU (gvctx->main_menu), NULL, NULL, NULL,
                            NULL, event_button->button, event_button->time);

        }
    }
}


void cb_dialog_abandonfile_save(GtkWidget *dialog,
                                gint response,
                                gpointer *data)
{
    outfile_dialog_ctx *ctx = (outfile_dialog_ctx *) data;

#ifdef _DEBUG
    printf("cb_dialog_abandonfile_save: called %s\n", *ctx->file);
#endif
    if (response == GTK_RESPONSE_OK) {
        ___put_default_abandoned_filename(*ctx->file);
    }
}



void cb_dialog_outputfile(GtkWidget *dialog, gint response, gpointer *data)
{
    outfile_dialog_ctx *ctx = (outfile_dialog_ctx *) data;
    gint fs_response;
    GtkWidget *fs;

#ifdef _DEBUG
    printf("cb_dialog_outputfile: Response code = %d\n", response);
#endif
    if (response == GTK_RESPONSE_YES) {
        fs = gtk_file_selection_new("Output file select");
        gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(fs));
        gtk_widget_show(fs);
        fs_response = gtk_dialog_run(GTK_DIALOG(fs));
        if (fs_response == GTK_RESPONSE_OK) {
            gtk_entry_set_text(GTK_ENTRY(ctx->entry),
                gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));
            response = GTK_RESPONSE_OK;
        }
        gtk_widget_destroy(fs);
        return;
    }
    gtk_widget_hide_all(ctx->dialog);
    if (response == GTK_RESPONSE_OK) {
        ctx->action = ctx->radio_selection;
        if (ctx->action == OUTPUT_FILE_APPEND_DEFAULT) {
            ctx->action = OUTPUT_FILE_APPEND_SPLIT;
        }

        if (*ctx->file) {
            free(*ctx->file);
        }
        *ctx->file = strdup((char *) gtk_entry_get_text(GTK_ENTRY(ctx->entry)));
    }
    else if (response == GTK_RESPONSE_REJECT) {
        gtk_entry_set_text(GTK_ENTRY(ctx->entry), *ctx->file);
    }
#ifdef _DEBUG
    printf("cb_dialog_outputfile: output_file=%s\n", ctx->gvctx->output_file);
    printf("cb_dialog_outputfile: abandon_file=%s\n", ctx->gvctx->abandon_file);
#endif
}


void cb_outfile_radio(GtkWidget *widget, gpointer data)
{
    outfile_radio_ctx *rbctx = (outfile_radio_ctx *) data;

    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
        *rbctx->radio_selection = rbctx->button;
    }
}


outfile_dialog_ctx *___make_dialog_outputfile(global_vars_ctx *gvctx,
                                              char *name_label,
                                              int  show_radio,
                                              char **output_name)
{
    GtkWidget *dialog;
    GtkWidget *box;
    GtkWidget *label;
    GtkWidget *radio;
    GtkWidget *entry;
    outfile_radio_ctx *rbctx;
    outfile_dialog_ctx *ctx;

    ctx = (outfile_dialog_ctx *) calloc(1, sizeof(*ctx));
    if (!ctx) {
        return NULL;
    }

    dialog = gtk_dialog_new_with_buttons(name_label,
                                         GTK_WINDOW(gvctx->main_window),
                                         GTK_DIALOG_MODAL, 
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
                                         GTK_STOCK_OPEN,   GTK_RESPONSE_YES,
                                         GTK_STOCK_OK,     GTK_RESPONSE_OK,
                                         NULL);
    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
    label = gtk_label_new(name_label);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
                       label, FALSE, FALSE, 0);


    /*
     * Text entry 
     */
    entry = gtk_entry_new();
    gtk_widget_show(entry);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), entry, TRUE, TRUE, 0);
    gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);


    if (show_radio) {
    /*
     * Radio buttons
     */
    box = gtk_vbox_new(FALSE, 0);
    radio = gtk_radio_button_new_with_label (NULL, "Join");
    gtk_box_pack_start (GTK_BOX (box), radio, TRUE, TRUE, 0);
        gtk_widget_show (radio);
    rbctx = (outfile_radio_ctx *) calloc(1, sizeof(*rbctx));
    rbctx->button = OUTPUT_FILE_APPEND_JOIN;
    rbctx->radio_selection = &ctx->radio_selection;
    g_signal_connect(G_OBJECT(radio), "toggled",
                     G_CALLBACK(cb_outfile_radio), rbctx);
    radio = gtk_radio_button_new_with_label_from_widget (
                 GTK_RADIO_BUTTON (radio), "Split");
    ctx->radio_split = radio;
    gtk_box_pack_start (GTK_BOX (box), radio, TRUE, TRUE, 0);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio), TRUE);
    gtk_widget_show (radio);
    rbctx = (outfile_radio_ctx *) calloc(1, sizeof(*rbctx));
    rbctx->button = OUTPUT_FILE_APPEND_SPLIT;
    rbctx->radio_selection = &ctx->radio_selection;
    g_signal_connect(G_OBJECT(radio), "toggled",
                     G_CALLBACK(cb_outfile_radio), rbctx);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), box, TRUE, TRUE, 0);
    }

    ctx->gvctx  = gvctx;
    ctx->file   = output_name;
    ctx->dialog = dialog;
    ctx->entry  = entry;
    return ctx;
}


/*
 * Populate other default values in the context
 */
void ___set_editor_defaults(global_vars_ctx *gvctx)
{
    if (gvctx->idlectx->play_file) {
        free(gvctx->idlectx->play_file);
    }
    gvctx->idlectx->play_file  = NULL;
    gvctx->eidlectx->play_file = NULL;
    gvctx->idlectx->dirty      = FALSE;
    gvctx->eidlectx->dirty     = FALSE;

    /*
     * Populate other default values in the context
     */
    if (gvctx->output_file) {
       free(gvctx->output_file);
    }
    gvctx->output_file = strdup("");
    gtk_entry_set_text(GTK_ENTRY(gvctx->ofdialogctx->entry),
                       gvctx->output_file);

    if (gvctx->abandon_file) {
        free(gvctx->abandon_file);
    }
    gvctx->abandon_file = ___get_default_abandoned_filename(NULL);

    gtk_entry_set_text(GTK_ENTRY(gvctx->abandon_file_ctx->entry),
                       gvctx->abandon_file);

    gtk_entry_set_text(GTK_ENTRY(gvctx->query_abandon_ctx->entry),
                       gvctx->abandon_file);
    gvctx->ofdialogctx->action = OUTPUT_FILE_APPEND_SPLIT;
    
    gtk_toggle_button_set_active(
        GTK_TOGGLE_BUTTON(gvctx->ofdialogctx->radio_split), TRUE);
}


void init_rc_file(char *file)
{
    char            **def_files;
    char            *cp;
    char            *rc_path = NULL;

    /*
     * Get list of Gtk rc files.  Assume last in list is
     * path to rc file in home directory.  When found, use this list
     * to build path to .xmpgedit-gtkrc file in home directory.
     */
    def_files = gtk_rc_get_default_files();
    if (def_files) {
        while (*(def_files+1)) {
            def_files++;
        }
        cp = strrchr(def_files[0], '/');
        if (cp) {
            rc_path = calloc(1, (cp-def_files[0]) +
                                sizeof(GTK_RC_FILE) + 1);
        }
        if (rc_path) {
            strncat(rc_path, def_files[0], cp-def_files[0]);
            strcat(rc_path, GTK_RC_FILE);
        }
        gtk_rc_add_default_file(rc_path);
        free(rc_path);
    }
}


int main(int argc, char *argv[])
{
    GtkWidget *window;
    GtkWidget *label;
    GtkWidget *label_trt;
    GtkWidget *box;
    GtkWidget *mainvbox;
    GtkWidget *mainhbox;
    GtkWidget *leftboxspacer;
    GtkWidget *rightboxspacer;
    GtkWidget *bottomboxspacer;
    GtkWidget *box_elapsed;
    GtkWidget *box_trt;
    GtkWidget *box_entry;
    GtkWidget *vbox_playback;
    GtkWidget *hbox_playback;
    GtkWidget *hbox3;
    GtkWidget *button;
    GtkWidget *separator;
    GtkWidget *entry;
    GtkWidget *arrow;
    GtkWidget *event_box;
    GtkWidget *menubar;
    GtkWidget *menubar_filemenu;
    GtkWidget *menubar_fileitem;
    GtkWidget *menubar_helpmenu;
    GtkWidget *menubar_helpitem;
    GtkWidget *menubar_editmenu;
    GtkWidget *menubar_edititem;
    idle_ctx        idlectx;
    idle_ctx        eidlectx;
    global_vars_ctx gvctx;
    long            play_sec = 0;
    long            play_usec = 0;
    char            *play_file = "";

    memset(&idlectx,  0, sizeof(idlectx));
    memset(&eidlectx, 0, sizeof(eidlectx));
    memset(&gvctx,    0, sizeof(gvctx));

    /*
     * Make gvctx and playback control contexts reference each other
     */
    idlectx.gvctx  = eidlectx.gvctx = &gvctx;
    gvctx.idlectx  = &idlectx;
    gvctx.eidlectx = &eidlectx;

    /*
     * Set configurable parameters in global context
     */
    read_config_file(CONFIG_FILE_NAME, &gvctx);

    init_rc_file(GTK_RC_FILE);
    gtk_init(&argc, &argv);
    gvctx.settings = gtk_settings_get_default();
    gvctx.tooltips = gtk_tooltips_new();
    gvctx.tooltips_entry = gtk_tooltips_new();
    gtk_tooltips_set_delay(gvctx.tooltips, 2000);
    gtk_tooltips_set_delay(gvctx.tooltips_entry, 2000);
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window),
                         XMPGEDIT_PRODUCT " " XMPGEDIT_VERSION);


    /*
     * Enable button press events on main window; events are processed by
     * cb_window_button_press().
     */
    gtk_widget_add_events(window, GDK_BUTTON_PRESS_MASK);
    g_signal_connect(G_OBJECT(window), "button_press_event",
                 G_CALLBACK(cb_window_button_press), &gvctx);

    g_signal_connect(G_OBJECT(window), "delete_event",
                     G_CALLBACK(cb_delete), &idlectx);
    g_signal_connect(G_OBJECT(window), "delete_event",
                     G_CALLBACK(cb_delete), &eidlectx);
    gvctx.main_window = window;

    mainvbox = gtk_vbox_new(FALSE, 0);
    mainhbox = gtk_hbox_new(FALSE, 0);
    box      = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(mainvbox);
    gtk_widget_show(mainhbox);
    gtk_widget_show(box);

    leftboxspacer = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(leftboxspacer);

    rightboxspacer = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(rightboxspacer);

    bottomboxspacer = gtk_hbox_new(FALSE, 0);
    gtk_widget_show(bottomboxspacer);

    gtk_container_add(GTK_CONTAINER(window), mainvbox);
    gtk_container_set_border_width(GTK_CONTAINER(box), 0);

    /*
     * Start packing main menu bar into top-level container.  The "file" menu
     * is populated below, when more required widgets have been constructed.
     */
    menubar = gtk_menu_bar_new();
    gtk_widget_show(menubar);
    gtk_box_pack_start(GTK_BOX(mainvbox), menubar, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(mainvbox), mainhbox, TRUE, TRUE, 0);

    gtk_box_pack_start(GTK_BOX(mainhbox), leftboxspacer, FALSE, FALSE, 5);
    gtk_box_pack_start(GTK_BOX(mainhbox), box, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(mainhbox), rightboxspacer, FALSE, FALSE, 5);

    gvctx.editctx.scrolled_window = 
        ___make_scrolled_window(&idlectx.gvctx->editctx.clist, &event_box);
    g_signal_connect(GTK_CLIST(gvctx.editctx.clist), "select_row",
                     G_CALLBACK(cb_row_clicked), &gvctx);
    g_signal_connect(GTK_CONTAINER(event_box), "button_press_event",
                     G_CALLBACK(cb_edit_menu), &gvctx);

    gtk_widget_set_size_request(gvctx.editctx.scrolled_window,
                                -1, 200);

    gtk_box_pack_start(GTK_BOX(box),
                       gvctx.editctx.scrolled_window, 
                       TRUE, TRUE, 1);

    /*
     * File name entry widgets
     */
    box_entry = gtk_hbox_new(FALSE, 5);
    gtk_widget_show(box_entry);
    gtk_box_pack_start(GTK_BOX(box), box_entry, FALSE, FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(box_entry), 3);

    /*
     * Elapsed, start, and TRT labels
     */
    box_elapsed = gtk_hbox_new(FALSE, 5);
    gtk_widget_show(box_elapsed);

    label = gtk_label_new("Elapsed Time:");
    gtk_widget_show(label);
    gtk_box_pack_start(GTK_BOX(box_elapsed), label, FALSE, FALSE, 0);
    label = gtk_label_new("elapsed_value");
    idlectx.label_elapsed = eidlectx.label_elapsed = label;
    gtk_widget_show(label);
    gtk_box_pack_start(GTK_BOX(box_elapsed), label, FALSE, FALSE, 0);
    label_set_text(idlectx.label_elapsed, "%d:%02d.%03d (%ds)    ", 0, 0, 0, 0);

    /*
     * TRT label
     */
    box_trt = gtk_hbox_new(FALSE, 5);
    gtk_widget_show(box_trt);

    label_trt = gtk_label_new("Total Time: 00:00.000 (0s)    ");
    gtk_widget_show(label_trt);
    idlectx.label_trt = eidlectx.label_trt = label_trt;
    gtk_box_pack_start(GTK_BOX(box_trt), label_trt, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(box_entry), box_elapsed, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(box_entry), box_trt, FALSE, FALSE, 20);
    
    button = gtk_toggle_button_new();
    gtk_widget_show(button);
    gvctx.open_file_toggle  = button;
    gtk_box_pack_end(GTK_BOX(box_entry), button, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(cb_open_file), &gvctx);

    entry = gtk_entry_new();
    gtk_tooltips_set_tip(gvctx.tooltips_entry, GTK_WIDGET(entry), 
                         play_file, NULL);
    gtk_entry_set_text(GTK_ENTRY(entry), play_file);
    gtk_editable_set_position(GTK_EDITABLE(entry), strlen(play_file));
    gtk_widget_show(entry);
    gtk_box_pack_end(GTK_BOX(box_entry), entry, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(entry), "activate",
                     G_CALLBACK(cb_entry), &gvctx);

    label = gtk_label_new("File: ");
    gtk_widget_show(label);
    gtk_box_pack_end(GTK_BOX(box_entry), label, FALSE, FALSE, 2);

    /*
     * Open button
     */
    gvctx.fs = gtk_file_selection_new ("File selection");
    gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(gvctx.fs));
    g_signal_connect(G_OBJECT(GTK_FILE_SELECTION(gvctx.fs)->ok_button), 
                     "clicked", G_CALLBACK(cb_file_sel_ok_1), &gvctx);

    g_signal_connect(G_OBJECT(GTK_FILE_SELECTION(gvctx.fs)->cancel_button),
                     "clicked", G_CALLBACK(cb_file_sel_cancel), &gvctx);


    arrow = gtk_arrow_new(GTK_ARROW_UP, GTK_SHADOW_OUT);
    gtk_widget_show(arrow);
    gtk_container_add(GTK_CONTAINER(button), arrow);

    gtk_entry_set_max_length(GTK_ENTRY(entry), PATH_MAX);
    gvctx.entry = entry;

    gvctx.volumectx = ___make_volume_control();
    gtk_widget_show_all(gvctx.volumectx->box);
    gtk_widget_hide(gvctx.volumectx->box);

    /*
     * Start playback controls population
     */
    vbox_playback = gtk_vbox_new(FALSE, 5);
    gtk_widget_show(vbox_playback);

    hbox_playback = gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox_playback);
    gtk_box_pack_start(GTK_BOX(box), hbox_playback, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(hbox_playback), vbox_playback, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(hbox_playback), gvctx.volumectx->box, 
                       FALSE, FALSE, 10);
                      
    separator = gtk_hseparator_new();
    gtk_widget_show(separator);
    gtk_box_pack_start(GTK_BOX(vbox_playback), separator, FALSE, FALSE, 0);

    __construct_playback_control(&idlectx, &gvctx, vbox_playback, 
                                 "Start Time:  ", play_sec, play_usec);

    separator = gtk_hseparator_new();
    gtk_widget_show(separator);
    gtk_box_pack_start(GTK_BOX(vbox_playback), separator, FALSE, FALSE, 0);

    __construct_playback_control(&eidlectx, &gvctx, vbox_playback, 
                                 "End Time:  ", play_sec, play_usec);

    separator = gtk_hseparator_new();
    gtk_widget_show(separator);
    gtk_box_pack_start(GTK_BOX(vbox_playback), separator, FALSE, FALSE, 0);
    /*
     * End of playback controls population
     */

    hbox3 = gtk_hbox_new(TRUE, 4);
    gtk_widget_show(hbox3);

    
    if (play_sec) {
        label_set_text(idlectx.label_trt, "Total Time: %d:%02d.%03d (%ds)", 
                       play_sec/60, play_sec%60, play_usec, play_sec);
    }
    gtk_box_pack_start(GTK_BOX(box), hbox3, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(box), bottomboxspacer, FALSE, FALSE, 5);

    /*
     * Populate the edit window with files named from the command line,
     * and create pop-up menu structures.
     */
    gvctx.main_menu = ___mainmenu_make(&gvctx);
    gvctx.edit_menu = ___editmenu_make(&gvctx);
    gvctx.ofdialogctx = ___make_dialog_outputfile(&gvctx,
                                                  "Output file name",
                                                  TRUE,
                                                  &gvctx.output_file);
    g_signal_connect(GTK_OBJECT (gvctx.ofdialogctx->dialog), 
                      "response", 
                      G_CALLBACK(cb_dialog_outputfile),
                      (gpointer) gvctx.ofdialogctx);

    gvctx.abandon_file_ctx = ___make_dialog_outputfile(&gvctx,
                                                  "Abandoned edits file name",
                                                  FALSE,
                                                  &gvctx.abandon_file);
    g_signal_connect(GTK_OBJECT (gvctx.abandon_file_ctx->dialog), 
                      "response", 
                      G_CALLBACK(cb_dialog_outputfile),
                      (gpointer) gvctx.abandon_file_ctx);
    g_signal_connect(GTK_OBJECT (gvctx.abandon_file_ctx->dialog),
                      "response", 
                      G_CALLBACK(cb_load),
                      (gpointer) &gvctx);

    gvctx.query_abandon_ctx = ___make_dialog_outputfile(
                                  &gvctx,
                                  "Abandoned file already exists, overwrite?",
                                  FALSE,
                                  &gvctx.abandon_file);
    g_signal_connect(GTK_OBJECT (gvctx.query_abandon_ctx->dialog), 
                      "response", 
                      G_CALLBACK(cb_dialog_outputfile),
                      (gpointer) gvctx.query_abandon_ctx);
    g_signal_connect(GTK_OBJECT (gvctx.query_abandon_ctx->dialog),
                      "response", 
                      G_CALLBACK(cb_edit_2),
                      (gpointer) &gvctx);

    ___set_editor_defaults(&gvctx);
    ___load_editor_blank_first_line(&gvctx);
    if (argc > 1) {
        ___load_editor_by_files(&gvctx, &argv[1], argc-1);
        gtk_clist_select_row(GTK_CLIST(gvctx.editctx.clist), argc-1, 0);
    }
    /*
     * file menu creation. adam: This is the same menu as
     * used for the popup.  I did not think this was allowed,
     * but it seems to work fine.
     */
    menubar_filemenu = gvctx.main_menu;
    menubar_fileitem = gtk_menu_item_new_with_label("File");
    gtk_widget_show(menubar_fileitem);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menubar_fileitem),
                              menubar_filemenu);
    gtk_menu_bar_append(GTK_MENU_BAR(menubar), menubar_fileitem);

    /*
     * Edit menu creation. adam: This is the same menu as
     * used for the popup.  I did not think this was allowed,
     * but it seems to work fine.
     */
    menubar_editmenu = gvctx.edit_menu;
    menubar_edititem = gtk_menu_item_new_with_label("Edit");
    gtk_widget_show(menubar_edititem);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menubar_edititem),
                              menubar_editmenu);
    gtk_menu_bar_append(GTK_MENU_BAR(menubar), menubar_edititem);

    /* Help menu creation */
    menubar_helpmenu = ___helpmenu_make(&gvctx);
    menubar_helpitem = gtk_menu_item_new_with_label("Help");
    gtk_widget_show(menubar_helpitem);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menubar_helpitem),
                              menubar_helpmenu);
    gtk_menu_bar_append(GTK_MENU_BAR(menubar), menubar_helpitem);
    gtk_menu_item_right_justify(GTK_MENU_ITEM(menubar_helpitem));

    /* About dialog creation */
    gvctx.about_dialog = ___aboutdialog_make(&gvctx);

    gtk_widget_show(window);
    gtk_main();
    return 0;
}

_MAIN(main)
